# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import abc
import datetime
import typing as t

import sspi.raw as raw

from ._filetime import filetime_to_datetime


class UnwrapResult(t.NamedTuple):
    data: bytes
    qop: int


class SecurityContext(raw.CtxtHandle, metaclass=abc.ABCMeta):
    """Base class for a SecurityContext."""

    def __init__(
        self,
        credential: raw.CredHandle | None = None,
        channel_bindings: raw.SecChannelBindings | None = None,
    ) -> None:
        self._complete = False
        self._credential = credential
        self._channel_bindings = channel_bindings
        self._expiry = datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)
        self._sizes = raw.SecPkgContextSizes()
        self.__seq_no = 0

    @property
    def expiry(self) -> datetime.datetime:
        """The time when this security context expires."""
        return self._expiry

    @property
    def complete(self) -> bool:
        """Whether this context has been authenticated and ready to use for message operations."""
        return self._complete

    @property
    def _seq_no(self) -> int:
        seq_no = self.__seq_no
        self.__seq_no += 1
        return seq_no

    def step(
        self,
        in_token: bytes | bytearray | memoryview | None = None,
    ) -> bytes:
        """Perform an authentication step.

        Performs and authentication step to authenticate the context. The
        in_token is the token received from the peer to use for the
        authentication step. The first step for the client typically has no
        in_token as it generates the first one.

        This method should be called until the context has been marked as
        complete.

        If the in_token is a bytearray or memoryview that is writable, SSPI
        can potentially change the bytes in the buffer. To avoid this provide
        a byte string or a readonly memoryview.

        Args:
            in_token: The input token to use for the authentication step.

        Returns:
            bytes: The token to send to the peer for it to use as the next
            step. This will be an empty byte string if no more tokens are
            available to send.

        Raises:
            WindowsError: If the authentication step failed the WindowsError
            contains more information around the failure.
        """
        sec_buffers = []
        if in_token:
            sec_buffers.append(
                self._get_secbuffer_data(in_token, raw.SecBufferType.SECBUFFER_TOKEN),
            )

        if self._channel_bindings:
            sec_buffers.append(self._channel_bindings.dangerous_get_sec_buffer())

        input_buffers = None
        if sec_buffers:
            input_buffers = raw.SecBufferDesc(sec_buffers)

        output_buffers = raw.SecBufferDesc(
            [
                raw.SecBuffer(None, raw.SecBufferType.SECBUFFER_TOKEN),
            ]
        )
        status, expiry = self._step_internal(input_buffers, output_buffers)

        if status == raw.NtStatus.SEC_E_OK:
            self._complete = True
            self._sizes = raw.query_context_attributes(self, raw.SecPkgContextSizes)

        self._expiry = filetime_to_datetime(expiry)

        return output_buffers[0].data

    def wrap(
        self,
        data: bytes | bytearray | memoryview,
        encrypt: bool = True,
    ) -> bytes:
        """Wraps the data provided.

        This method is used to wrap the data in order to provide confidentially
        and integrity to the message. This method is designed to create a
        wrapped result that can be used by the :meth:`unwrap` function as well
        as GSSAPI's ``gss_unwrap`` function.

        If the provided data was a bytearray or a writable memoryview SSPI can
        mutate the data during the wrapping operation. If this is not desired
        then provide a byte string or readonly memoryview.

        This function defaults to encrypting the data. Set ``encrypt=False`` if
        you wish to generate a signature or use the :meth:`sign` method. It is
        up to the security provider that was negotiated in the authentication
        stage to honour the encryption request.

        If a more complex set of buffers is needed to wrap the data, use the
        :meth:`sspi.raw.encrypt_message` function with this context.

        Args:
            data: The data to wrap.
            encrypt: Whether to encrypt the data or just sign it.

        Returns:
            bytes: The wrapped data.

        Raises:
            WindowsError: Contains the error information if the wrap failed.
        """
        token_data = bytearray(self._sizes.security_trailer)
        padding_data = bytearray(self._sizes.block_size)
        buffer = raw.SecBufferDesc(
            [
                raw.SecBuffer(token_data, raw.SecBufferType.SECBUFFER_TOKEN),
                self._get_secbuffer_data(data),
                raw.SecBuffer(padding_data, raw.SecBufferType.SECBUFFER_PADDING),
            ],
        )

        raw.encrypt_message(
            context=self,
            qop=0 if encrypt else raw.QopFlags.SECQOP_WRAP_NO_ENCRYPT,
            message=buffer,
            seq_no=self._seq_no,
        )

        return buffer[0].data + buffer[1].data + buffer[2].data

    def unwrap(
        self,
        data: bytes | bytearray | memoryview,
    ) -> UnwrapResult:
        """Unwraps the data provided.

        This method is used to unwrap the data provided to verify its integrity
        as well as decrypt the value. This method is designed to be able to
        unwrap a result from :meth:`wrap` as well as GSSAPI's ``gss_wrap``
        function.

        If the provided data was a bytearray or a writable memoryview SSPI can
        mutate the data during the unwrapping operation. If this is not desired
        then provide a byte string or readonly memoryview.

        If a more complex set of buffers is needed to unwrap the data, use the
        :meth:`sspi.raw.decrypt_message` function with this context.

        Args:
            data: The data to unwrap.

        Returns:
            UnwrapResult: A tuple containing the unwrapped data and QoP value
            returned by the security provider for this operation.

        Raises:
            WindowsError: Contains the error information if the unwrap failed.
        """
        buffer = raw.SecBufferDesc(
            [
                self._get_secbuffer_data(data, raw.SecBufferType.SECBUFFER_STREAM),
                raw.SecBuffer(None, raw.SecBufferType.SECBUFFER_DATA),
            ]
        )
        qop = raw.decrypt_message(
            context=self,
            message=buffer,
            seq_no=self._seq_no,
        )

        return UnwrapResult(
            data=buffer[1].data,
            qop=qop,
        )

    def sign(
        self,
        data: bytes | bytearray | memoryview,
    ) -> bytes:
        """Signs the data provided.

        This method is used to sign the data in order to provide integrity to
        the message.

        If the provided data was a bytearray or a writable memoryview SSPI can
        mutate the data during the signing operation. If this is not desired
        then provide a byte string or readonly memoryview.

        If a more complex set of buffers is needed to sign the data, use the
        :meth:`sspi.raw.make_signature` function with this context.

        Args:
            data: The data to sign.

        Returns:
            bytes: The signature for the data provided.

        Raises:
            WindowsError: Contains the error information if the signature
            operation failed.
        """
        in_token = bytearray(self._sizes.max_signature)
        buffer = raw.SecBufferDesc(
            [
                self._get_secbuffer_data(data),
                raw.SecBuffer(in_token, raw.SecBufferType.SECBUFFER_TOKEN),
            ]
        )

        raw.make_signature(
            context=self,
            qop=0,
            message=buffer,
            seq_no=self._seq_no,
        )

        return buffer[1].data

    def verify(
        self,
        data: bytes | bytearray | memoryview,
        mic: bytes | bytearray | memoryview,
    ) -> int:
        """Verifies the data provided.

        This method is used to verify a signature to ensure the data has not
        been modified.

        If the provided data was a bytearray or a writable memoryview SSPI can
        mutate the data during the unwrapping operation. If this is not desired
        then provide a byte string or readonly memoryview.

        If a more complex set of buffers is needed to unwrap the data, use the
        :meth:`sspi.raw.verify_signature` function with this context.

        Args:
            data: The data to verify.
            mic: The signature to verify with

        Returns:
            int: The QoP value that was applied to the signature.

        Raises:
            WindowsError: Contains the error information if the verify
            operation failed.
        """
        buffer = raw.SecBufferDesc(
            [
                self._get_secbuffer_data(data),
                self._get_secbuffer_data(mic, raw.SecBufferType.SECBUFFER_TOKEN),
            ]
        )

        return raw.verify_signature(
            context=self,
            message=buffer,
            seq_no=self._seq_no,
        )

    @abc.abstractmethod
    def _step_internal(
        self,
        input_buffers: raw.SecBufferDesc | None,
        output_buffers: raw.SecBufferDesc | None,
    ) -> tuple[int, int]:
        ...

    def _get_secbuffer_data(
        self,
        data: bytes | bytearray | memoryview,
        buffer_type: raw.SecBufferType = raw.SecBufferType.SECBUFFER_DATA,
    ) -> raw.SecBuffer:
        buffer: bytearray | memoryview
        if isinstance(data, bytes) or (isinstance(data, memoryview) and data.readonly):
            buffer = bytearray(data)
        else:
            buffer = data

        return raw.SecBuffer(buffer, buffer_type)


class ClientSecurityContext(SecurityContext):
    """A client security context.

    This represents an SSPI security context that can be used for client side
    authentication. This class is designed to be a high level overlay on top of
    the :class:`sspi.raw.CtxtHandle` class. it can be used directly with
    any low level API in the ``sspi.raw`` namespace that requires a context
    handle instance in case the low level interface doesn't expose the methods
    needed.

    The client must specify a target_name which is typically the Service
    Principal Name (SPN) of the service. It might be a different value
    depending on the security protocol used.

    By default the flags are set to request mutual auth, replay detection,
    sequence detection, confidentiality, and integrity if no flags are
    specified. Custom flags can be requested but it will overwrite the default
    set.

    The credential stores the credential to use for authentication as well as
    extra information that controls how the security protocol will perform
    the authentication steps.

    Optional channel bindings can be specified which is used to tie the
    security context with an outer channel.

    Properties like ``expiry`` and ``attributes`` won't be populated until the
    context has been negotiated and is complete.

    Args:
        target_name: The target service's name, typically the SPN.
        flags: Custom ISC REQ flags to use.
        credential: A credential to use for authentication, defaults to the
            current user context if not set.
        channel_bindings: Optional channel bindings to tie the context to.
    """

    def __init__(
        self,
        target_name: str,
        flags: raw.IscReq | int | None = None,
        credential: raw.CredHandle | None = None,
        *,
        channel_bindings: raw.SecChannelBindings | None = None,
    ) -> None:
        super().__init__(credential=credential, channel_bindings=channel_bindings)
        self._target_name = target_name
        self._flags = (
            flags
            if flags is not None
            else (
                raw.IscReq.ISC_REQ_MUTUAL_AUTH
                | raw.IscReq.ISC_REQ_REPLAY_DETECT
                | raw.IscReq.ISC_REQ_SEQUENCE_DETECT
                | raw.IscReq.ISC_REQ_CONFIDENTIALITY
                | raw.IscReq.ISC_REQ_INTEGRITY
            )
        )
        self._attributes: raw.IscRet = raw.IscRet(0)

    @property
    def attributes(self) -> raw.IscRet:
        """Attributes of the current security context."""
        return self._attributes

    def _step_internal(
        self,
        input_buffers: raw.SecBufferDesc | None,
        output_buffers: raw.SecBufferDesc | None,
    ) -> tuple[int, int]:
        result = raw.initialize_security_context(
            credential=self._credential,
            context=self,
            target_name=self._target_name,
            context_req=self._flags | raw.IscReq.ISC_REQ_ALLOCATE_MEMORY,
            target_data_rep=raw.TargetDataRep.SECURITY_NATIVE_DREP,
            input_buffers=input_buffers,
            output_buffers=output_buffers,
        )
        self._attributes = result.attributes

        return result.status, result.expiry


class ServerSecurityContext(SecurityContext):
    """A server security context.

    This represents an SSPI security context that can be used for server side
    authentication. This class is designed to be a high level overlay on top of
    the :class:`sspi.raw.CtxtHandle` class. it can be used directly with
    any low level API in the ``sspi.raw`` namespace that requires a context
    handle instance in case the low level interface doesn't expose the methods
    needed.

    By default the flags are set to request mutual auth, replay detection,
    sequence detection, confidentiality, and integrity if no flags are
    specified. Custom flags can be requested but it will overwrite the default
    set.

    The credential stores the credential to use for authentication as well as
    extra information that controls how the security protocol will perform
    the authentication steps.

    Optional channel bindings can be specified which is used to tie the
    security context with an outer channel.

    Properties like ``expiry`` and ``attributes`` won't be populated until the
    context has been negotiated and is complete.

    Args:
        flags: Custom ASC REQ flags to use.
        credential: A credential to use for authentication
        channel_bindings: Optional channel bindings to tie the context to.
    """

    def __init__(
        self,
        flags: raw.AscReq | int | None = None,
        credential: raw.CredHandle | None = None,
        *,
        channel_bindings: raw.SecChannelBindings | None = None,
    ) -> None:
        super().__init__(credential=credential, channel_bindings=channel_bindings)
        self._flags = (
            flags
            if flags is not None
            else (
                raw.AscReq.ASC_REQ_MUTUAL_AUTH
                | raw.AscReq.ASC_REQ_REPLAY_DETECT
                | raw.AscReq.ASC_REQ_SEQUENCE_DETECT
                | raw.AscReq.ASC_REQ_CONFIDENTIALITY
                | raw.AscReq.ASC_REQ_INTEGRITY
            )
        )
        self._attributes: raw.AscRet = raw.AscRet(0)

    @property
    def attributes(self) -> raw.AscRet:
        """Attributes of the current security context."""
        return self._attributes

    def _step_internal(
        self,
        input_buffers: raw.SecBufferDesc | None,
        output_buffers: raw.SecBufferDesc | None,
    ) -> tuple[int, int]:
        result = raw.accept_security_context(
            credential=self._credential,
            context=self,
            input_buffers=input_buffers,
            context_req=self._flags | raw.AscReq.ASC_REQ_ALLOCATE_MEMORY,
            target_data_rep=raw.TargetDataRep.SECURITY_NATIVE_DREP,
            output_buffers=output_buffers,
        )
        self._attributes = result.attributes

        status = result.status

        # sspi-rs is weird for NTLM where it returns SEC_I_COMPLETE_NEEDED.
        # Until that bug is fixed and we pull in a new version we need to
        # manually handle that here.
        # https://github.com/Devolutions/sspi-rs/issues/167
        if status == raw.NtStatus.SEC_I_COMPLETE_NEEDED:
            raw.complete_auth_token(self, input_buffers or raw.SecBufferDesc([]))
            status = raw.NtStatus.SEC_E_OK

        return status, result.expiry



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from ._win32_types cimport *


cdef class WideCharString:
    cdef LPWSTR raw
    cdef Py_ssize_t length


cdef str wide_char_to_str(
    const LPWSTR value,
    int size = *,
    int none_is_empty = *,
)



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

from cpython.mem cimport PyMem_Free
from cpython.ref cimport PyObject


cdef extern from "python_sspi.h":
    # Linux defines wchar_t as UTF-32 whereas the APIs expect UTF-16. This
    # tries to paper over the differences with a shim.
    """
    #if defined(PYSSPI_IS_LINUX)

    #include <unicode/ustring.h>

    LPWSTR PyUnicode_AsWideCharStringShim(PyObject * unicode, Py_ssize_t *size)
    {
        Py_ssize_t utf32_size = 0;
        wchar_t * utf32_buffer = PyUnicode_AsWideCharString(unicode, &utf32_size);

        UChar * utf16_buffer = NULL;
        int32_t utf16_capacity = 0;
        while (1)
        {
            int32_t utf16_chars = 0;
            UErrorCode error_code = 0;
            u_strFromUTF32(
                utf16_buffer,
                utf16_capacity,
                &utf16_chars,
                utf32_buffer,
                utf32_size,
                &error_code
            );

            if (error_code == U_BUFFER_OVERFLOW_ERROR)
            {
                // Ensure enough length for the NULL terminator
                utf16_capacity = utf16_chars + 1;
                utf16_buffer = realloc(utf16_buffer, utf16_capacity * 2);
                continue;
            }
            else if (error_code > 0)
            {
                if (utf16_buffer != NULL)
                {
                    free(utf16_buffer);
                }
                utf16_buffer = NULL;
                break;
            }
            else
            {
                *size = utf16_chars;
                break;
            }
        }

        PyMem_Free(utf32_buffer);
        return utf16_buffer;
    }

    PyObject *PyUnicode_FromWideCharShim(const LPWSTR w, Py_ssize_t size)
    {
        // Get the number of UTF-32 codepoints needed for the input string.
        UErrorCode error_code = 0;
        int32_t utf32_size = 0;
        u_strToUTF32(
            NULL,
            0,
            &utf32_size,
            w,
            size,
            &error_code
        );
        if (error_code != U_BUFFER_OVERFLOW_ERROR)
        {
            return NULL;
        }

        // Allocate the UTF-32 buffer including the NULL terminator. As it's a
        // fixed size we can just do 4 bytes per codepoint.
        UChar32* utf32_buffer = (UChar32*)malloc((utf32_size + 1) * 4);
        if (utf32_buffer == NULL)
        {
            return NULL;
        }

        error_code = 0;
        u_strToUTF32(
            utf32_buffer,
            utf32_size + 1,
            NULL,
            w,
            size,
            &error_code
        );

        PyObject * py_str = NULL;
        if (error_code <= 0)
        {
            py_str = PyUnicode_FromWideChar(utf32_buffer, utf32_size);
        }
        free(utf32_buffer);

        return py_str;
    }

    void PyMem_FreeShim(void *p)
    {
        free(p);
    }

    #else

    LPWSTR PyUnicode_AsWideCharStringShim(PyObject * unicode, Py_ssize_t *size)
    {
        return PyUnicode_AsWideCharString(unicode, size);
    }

    PyObject *PyUnicode_FromWideCharShim(const LPWSTR w, Py_ssize_t size)
    {
        return PyUnicode_FromWideChar(w, size);
    }

    void PyMem_FreeShim(void *p)
    {
        PyMem_Free(p);
    }

    #endif
    """

    LPWSTR PyUnicode_AsWideCharStringShim(object unicode, Py_ssize_t *size)
    PyObject *PyUnicode_FromWideCharShim(const LPWSTR w, Py_ssize_t size)
    void PyMem_FreeShim(void *p)


cdef class WideCharString:
    # cdef LPWSTR raw
    # cdef Py_ssize_t length

    def __cinit__(WideCharString self, str value) -> None:
        if value is not None:
            self.raw = PyUnicode_AsWideCharStringShim(value, &self.length)
            if self.raw == NULL:
                raise MemoryError()

    def __dealloc__(WideCharString self) -> None:
        if self.raw != NULL:
            PyMem_FreeShim(self.raw)

        self.raw = NULL
        self.length = 0

cdef str wide_char_to_str(
    const LPWSTR value,
    int size = -1,
    int none_is_empty = 0,
):
    if value == NULL:
        return "" if none_is_empty else None

    return <object>PyUnicode_FromWideCharShim(value, size)



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import datetime
import os

import pytest

import sspi


def test_sec_context_attributes(
    initial_contexts: tuple[sspi.ClientSecurityContext, sspi.ServerSecurityContext],
) -> None:
    c_ctx, s_ctx = initial_contexts

    assert not c_ctx.complete
    assert c_ctx.expiry == datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)
    assert c_ctx.attributes == sspi.IscRet(0)

    assert not s_ctx.complete
    assert s_ctx.expiry == datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)
    assert s_ctx.attributes == sspi.AscRet(0)

    s_token = None
    while not (c_ctx.complete and not s_token):
        c_token = c_ctx.step(s_token)

        if c_token:
            s_token = s_ctx.step(c_token)
        else:
            s_token = None

    assert c_ctx.complete
    assert c_ctx.expiry != datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)
    assert c_ctx.attributes != sspi.IscRet(0)

    assert s_ctx.complete
    assert isinstance(s_ctx.expiry, datetime.datetime)
    assert s_ctx.attributes != sspi.AscRet(0)


# https://github.com/Devolutions/sspi-rs/issues/84
@pytest.mark.skipif(os.name != "nt", reason="sspi-rs STREAM unwrapping does not support NTLM yet")
def test_sec_context_wrapping(
    authenticated_contexts: tuple[sspi.ClientSecurityContext, sspi.ServerSecurityContext],
) -> None:
    client_data = os.urandom(32)
    server_data = os.urandom(32)
    c_ctx, s_ctx = authenticated_contexts

    wrapped_data = c_ctx.wrap(client_data)
    assert wrapped_data != client_data

    unwrapped_data = s_ctx.unwrap(wrapped_data)
    assert isinstance(unwrapped_data, sspi.UnwrapResult)
    assert unwrapped_data.qop == 0
    assert unwrapped_data.data == client_data

    wrapped_data = s_ctx.wrap(server_data)
    assert wrapped_data != server_data

    unwrapped_data = c_ctx.unwrap(wrapped_data)
    assert isinstance(unwrapped_data, sspi.UnwrapResult)
    assert unwrapped_data.qop == 0
    assert unwrapped_data.data == server_data


@pytest.mark.skipif(os.name != "nt", reason="sspi-rs does not support signatures yet")
def test_sec_context_signatures(
    authenticated_contexts: tuple[sspi.ClientSecurityContext, sspi.ServerSecurityContext],
) -> None:
    client_data = os.urandom(32)
    server_data = os.urandom(32)
    c_ctx, s_ctx = authenticated_contexts

    c_signature = c_ctx.sign(client_data)
    s_ctx.verify(client_data, c_signature)

    s_signature = s_ctx.sign(server_data)
    c_ctx.verify(server_data, s_signature)


def test_wrap_with_bytearray(
    authenticated_contexts: tuple[sspi.ClientSecurityContext, sspi.ServerSecurityContext],
) -> None:
    client_data = os.urandom(32)
    b_client_data = bytearray(client_data)
    c_ctx, s_ctx = authenticated_contexts

    wrapped_data = c_ctx.wrap(b_client_data)
    assert bytes(b_client_data) != client_data  # See if it's mutated in place
    assert wrapped_data != client_data

    if os.name != "nt":
        return

    b_wrapped_data = bytearray(wrapped_data)
    unwrapped_data = s_ctx.unwrap(b_wrapped_data)
    assert isinstance(unwrapped_data, sspi.UnwrapResult)
    assert unwrapped_data.qop == 0
    assert unwrapped_data.data == client_data
    assert bytes(b_wrapped_data) != wrapped_data


def test_wrap_with_writable_memoryview(
    authenticated_contexts: tuple[sspi.ClientSecurityContext, sspi.ServerSecurityContext],
) -> None:
    client_data = os.urandom(32)
    b_client_data = memoryview(bytearray(client_data))
    c_ctx, s_ctx = authenticated_contexts

    wrapped_data = c_ctx.wrap(b_client_data)
    assert bytes(b_client_data) != client_data  # See if it's mutated in place
    assert wrapped_data != client_data

    if os.name != "nt":
        return

    b_wrapped_data = memoryview(bytearray(wrapped_data))
    unwrapped_data = s_ctx.unwrap(b_wrapped_data)
    assert isinstance(unwrapped_data, sspi.UnwrapResult)
    assert unwrapped_data.qop == 0
    assert unwrapped_data.data == client_data
    assert bytes(b_wrapped_data) != wrapped_data


def test_wrap_with_readonly_memoryview(
    authenticated_contexts: tuple[sspi.ClientSecurityContext, sspi.ServerSecurityContext],
) -> None:
    client_data = os.urandom(32)
    b_client_data = memoryview(client_data)
    c_ctx, s_ctx = authenticated_contexts

    wrapped_data = c_ctx.wrap(b_client_data)
    assert bytes(b_client_data) == client_data
    assert wrapped_data != client_data

    if os.name != "nt":
        return

    b_wrapped_data = memoryview(wrapped_data)
    unwrapped_data = s_ctx.unwrap(b_wrapped_data)
    assert isinstance(unwrapped_data, sspi.UnwrapResult)
    assert unwrapped_data.qop == 0
    assert unwrapped_data.data == client_data
    assert bytes(b_wrapped_data) == wrapped_data



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import collections
import enum

from libc.stdint cimport uint64_t

from ._text cimport WideCharString, wide_char_to_str
from ._win32_types cimport *


cdef extern from "python_sspi.h":
    unsigned int _SECPKG_CRED_INBOUND "SECPKG_CRED_INBOUND"
    unsigned int _SECPKG_CRED_OUTBOUND "SECPKG_CRED_OUTBOUND"
    unsigned int _SECPKG_CRED_BOTH "SECPKG_CRED_BOTH"
    unsigned int _SECPKG_CRED_AUTOLOGON_RESTRICTED "SECPKG_CRED_AUTOLOGON_RESTRICTED"
    unsigned int _SECPKG_CRED_PROCESS_POLICY_ONLY "SECPKG_CRED_PROCESS_POLICY_ONLY"

    unsigned int _SEC_WINNT_AUTH_IDENTITY_ANSI "SEC_WINNT_AUTH_IDENTITY_ANSI"
    unsigned int _SEC_WINNT_AUTH_IDENTITY_UNICODE "SEC_WINNT_AUTH_IDENTITY_UNICODE"
    unsigned int _SEC_WINNT_AUTH_IDENTITY_MARSHALLED "SEC_WINNT_AUTH_IDENTITY_MARSHALLED"
    unsigned int _SEC_WINNT_AUTH_IDENTITY_ONLY "SEC_WINNT_AUTH_IDENTITY_ONLY"
    unsigned int _SEC_WINNT_AUTH_IDENTITY_FLAGS_PROCESS_ENCRYPTED "SEC_WINNT_AUTH_IDENTITY_FLAGS_PROCESS_ENCRYPTED"
    unsigned int _SEC_WINNT_AUTH_IDENTITY_FLAGS_SYSTEM_PROTECTED "SEC_WINNT_AUTH_IDENTITY_FLAGS_SYSTEM_PROTECTED"
    unsigned int _SEC_WINNT_AUTH_IDENTITY_FLAGS_USER_PROTECTED "SEC_WINNT_AUTH_IDENTITY_FLAGS_USER_PROTECTED"
    unsigned int _SEC_WINNT_AUTH_IDENTITY_FLAGS_SYSTEM_ENCRYPTED "SEC_WINNT_AUTH_IDENTITY_FLAGS_SYSTEM_ENCRYPTED"
    unsigned int _SEC_WINNT_AUTH_IDENTITY_FLAGS_RESERVED "SEC_WINNT_AUTH_IDENTITY_FLAGS_RESERVED"
    unsigned int _SEC_WINNT_AUTH_IDENTITY_FLAGS_NULL_USER "SEC_WINNT_AUTH_IDENTITY_FLAGS_NULL_USER"
    unsigned int _SEC_WINNT_AUTH_IDENTITY_FLAGS_NULL_DOMAIN "SEC_WINNT_AUTH_IDENTITY_FLAGS_NULL_DOMAIN"
    unsigned int _SEC_WINNT_AUTH_IDENTITY_FLAGS_ID_PROVIDER "SEC_WINNT_AUTH_IDENTITY_FLAGS_ID_PROVIDER"

    unsigned int SEC_WINNT_AUTH_IDENTITY_VERSION
    unsigned int SEC_WINNT_AUTH_IDENTITY_VERSION_2

    cdef struct _SEC_WINNT_AUTH_IDENTITY_EXW:
        unsigned int Version;
        unsigned int Length;
        unsigned short *User;
        unsigned int UserLength;
        unsigned short *Domain;
        unsigned int DomainLength;
        unsigned short *Password;
        unsigned int PasswordLength;
        unsigned int Flags;
        unsigned short *PackageList;
        unsigned int PackageListLength;
    ctypedef _SEC_WINNT_AUTH_IDENTITY_EXW SEC_WINNT_AUTH_IDENTITY_EXW
    ctypedef SEC_WINNT_AUTH_IDENTITY_EXW *PSEC_WINNT_AUTH_IDENTITY_EXW

    ctypedef void (*SEC_GET_KEY_FN)(
        void *Arg,
        void *Principal,
        unsigned int KeyVer,
        void **Key,
        SECURITY_STATUS *Status,
    )

    # https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-acquirecredentialshandlew
    SECURITY_STATUS AcquireCredentialsHandleW(
        LPWSTR         pPrincipal,
        LPWSTR         pPackage,
        unsigned int  fCredentialUse,
        void           *pvLogonId,
        void           *pAuthData,
        SEC_GET_KEY_FN pGetKeyFn,
        void           *pvGetKeyArgument,
        PCredHandle    phCredential,
        PTimeStamp     ptsExpiry
    ) nogil

    # https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-freecredentialshandle
    SECURITY_STATUS FreeCredentialsHandle(
        PCredHandle phCredential
    ) nogil

class CredentialUse(enum.IntFlag):
    SECPKG_CRED_INBOUND = _SECPKG_CRED_INBOUND
    SECPKG_CRED_OUTBOUND = _SECPKG_CRED_OUTBOUND
    SECPKG_CRED_BOTH = _SECPKG_CRED_BOTH
    SECPKG_CRED_AUTOLOGON_RESTRICTED = _SECPKG_CRED_AUTOLOGON_RESTRICTED
    SECPKG_CRED_PROCESS_POLICY_ONLY = _SECPKG_CRED_PROCESS_POLICY_ONLY

class WinNTAuthFlags(enum.IntFlag):
    SEC_WINNT_AUTH_IDENTITY_ANSI = _SEC_WINNT_AUTH_IDENTITY_ANSI
    SEC_WINNT_AUTH_IDENTITY_UNICODE = _SEC_WINNT_AUTH_IDENTITY_UNICODE
    SEC_WINNT_AUTH_IDENTITY_MARSHALLED = _SEC_WINNT_AUTH_IDENTITY_MARSHALLED
    SEC_WINNT_AUTH_IDENTITY_ONLY = _SEC_WINNT_AUTH_IDENTITY_ONLY
    SEC_WINNT_AUTH_IDENTITY_FLAGS_PROCESS_ENCRYPTED = _SEC_WINNT_AUTH_IDENTITY_FLAGS_PROCESS_ENCRYPTED
    SEC_WINNT_AUTH_IDENTITY_FLAGS_SYSTEM_PROTECTED = _SEC_WINNT_AUTH_IDENTITY_FLAGS_SYSTEM_PROTECTED
    SEC_WINNT_AUTH_IDENTITY_FLAGS_USER_PROTECTED = _SEC_WINNT_AUTH_IDENTITY_FLAGS_USER_PROTECTED
    SEC_WINNT_AUTH_IDENTITY_FLAGS_SYSTEM_ENCRYPTED = _SEC_WINNT_AUTH_IDENTITY_FLAGS_SYSTEM_ENCRYPTED
    SEC_WINNT_AUTH_IDENTITY_FLAGS_RESERVED = _SEC_WINNT_AUTH_IDENTITY_FLAGS_RESERVED
    SEC_WINNT_AUTH_IDENTITY_FLAGS_NULL_USER = _SEC_WINNT_AUTH_IDENTITY_FLAGS_NULL_USER
    SEC_WINNT_AUTH_IDENTITY_FLAGS_NULL_DOMAIN = _SEC_WINNT_AUTH_IDENTITY_FLAGS_NULL_DOMAIN
    SEC_WINNT_AUTH_IDENTITY_FLAGS_ID_PROVIDER = _SEC_WINNT_AUTH_IDENTITY_FLAGS_ID_PROVIDER

cdef class CredHandle:
    # cdef _CredHandle raw
    # cdef int needs_free

    def __dealloc__(CredHandle self):
        if self.needs_free:
            FreeCredentialsHandle(&self.raw)
            self.needs_free = 0

cdef class AuthIdentity:

    cdef void *__c_value__(AuthIdentity self):
        return NULL

cdef class WinNTAuthIdentity(AuthIdentity):
    cdef SEC_WINNT_AUTH_IDENTITY_EXW raw
    cdef WideCharString username_wchar
    cdef WideCharString domain_wchar
    cdef WideCharString password_wchar
    cdef WideCharString package_list_wchar

    def __cinit__(
        WinNTAuthIdentity self,
        username: str | None = None,
        domain: str | None = None,
        password: str | None = None,
        *,
        flags: WinNTAuthFlags | int = 0,
        package_list: str | None = None,
    ) -> None:
        self.raw.Version = SEC_WINNT_AUTH_IDENTITY_VERSION
        self.raw.Length = <unsigned int>sizeof(SEC_WINNT_AUTH_IDENTITY_EXW)

        self.username_wchar = WideCharString(username)
        self.raw.User = <unsigned short*>self.username_wchar.raw
        self.raw.UserLength = <unsigned int>self.username_wchar.length

        self.domain_wchar = WideCharString(domain)
        self.raw.Domain = <unsigned short*>self.domain_wchar.raw
        self.raw.DomainLength = <unsigned int>self.domain_wchar.length

        self.password_wchar = WideCharString(password)
        self.raw.Password = <unsigned short*>self.password_wchar.raw
        self.raw.PasswordLength = <unsigned int>self.password_wchar.length

        self.raw.Flags = int(flags) | _SEC_WINNT_AUTH_IDENTITY_UNICODE

        self.package_list_wchar = WideCharString(package_list)
        self.raw.PackageList = <unsigned short*>self.package_list_wchar.raw
        self.raw.PackageListLength = <unsigned int>self.package_list_wchar.length

    cdef void *__c_value__(WinNTAuthIdentity self):
        return &self.raw

    def __repr__(WinNTAuthIdentity self) -> str:
        kwargs = [f"{k}={v}" for k, v in {
            'username': repr(self.username),
            'domain': repr(self.domain),
            'password': repr(self.password),
            'flags': self.flags,
            'package_list': repr(self.package_list),
        }.items()]

        return f"WinNTAuthIdentity({', '.join(kwargs)})"

    def __str__(WinNTAuthIdentity self) -> str:
        username = self.username
        if self.domain:
            username = f"{self.domain}\\{username}"

        return username

    @property
    def username(self) -> str | None:
        return wide_char_to_str(<LPWSTR>self.raw.User, self.raw.UserLength)

    @property
    def domain(self) -> str | None:
        return wide_char_to_str(<LPWSTR>self.raw.Domain, self.raw.DomainLength)

    @property
    def password(self) -> str | None:
        return wide_char_to_str(<LPWSTR>self.raw.Password, self.raw.PasswordLength)

    @property
    def flags(self) -> WinNTAuthFlags:
        return WinNTAuthFlags(self.raw.Flags)

    @property
    def package_list(self) -> str | None:
        return wide_char_to_str(<LPWSTR>self.raw.PackageList, self.raw.PackageListLength)

AcquireCredentialsResult = collections.namedtuple(
    'AcquireCredentialsResult',
    ['credential', 'expiry'],
)

def acquire_credentials_handle(
    str principal,
    str package not None,
    unsigned int credential_use,
    *,
    AuthIdentity auth_data = None,
) -> AcquireCredentialsResult:
    cdef CredHandle cred = CredHandle()
    cdef WideCharString principal_wstr = WideCharString(principal)
    cdef WideCharString package_wstr = WideCharString(package)
    cdef void *auth_data_buffer = NULL
    cdef TimeStamp raw_expiry

    if auth_data is not None:
        auth_data_buffer = auth_data.__c_value__()

    with nogil:
        res = AcquireCredentialsHandleW(
            principal_wstr.raw,
            package_wstr.raw,
            credential_use,
            NULL,
            auth_data_buffer,
            NULL,
            NULL,
            &cred.raw,
            &raw_expiry
        )

    if res != 0:
        PyErr_SetFromWindowsErr(res)

    cred.needs_free = 1

    return AcquireCredentialsResult(
        credential=cred,
        expiry=(<uint64_t>raw_expiry.HighPart << 32) | raw_expiry.LowPart,
    )

def _replace_cred_handle(
    CredHandle src not None,
    CredHandle dst not None,
) -> None:
    # This is only used by sspi._credential.py to store the cred state in
    # itself.
    dst.raw = src.raw
    dst.needs_free = src.needs_free
    src.needs_free = 0



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import os
import socket

import pytest

import sspi


@pytest.fixture()
def initial_contexts() -> tuple[sspi.ClientSecurityContext, sspi.ServerSecurityContext]:
    spn = f"host/{socket.gethostname()}"

    # sspi-rs only supports acceptors for NTLM at this point in time. it also
    # cannot rely on implicit creds
    if os.name == "nt":
        c_cred = sspi.UserCredential(protocol="NTLM")
        s_cred = sspi.UserCredential(usage="accept")
    else:
        c_cred = sspi.UserCredential("user", "pass", protocol="NTLM")
        s_cred = sspi.UserCredential("user", "pass", protocol="NTLM", usage="accept")

    c_ctx = sspi.ClientSecurityContext(target_name=spn, credential=c_cred)
    s_ctx = sspi.ServerSecurityContext(credential=s_cred)

    return c_ctx, s_ctx


@pytest.fixture()
def authenticated_contexts(
    initial_contexts: tuple[sspi.ClientSecurityContext, sspi.ServerSecurityContext],
) -> tuple[sspi.ClientSecurityContext, sspi.ServerSecurityContext]:
    c_ctx, s_ctx = initial_contexts

    s_token = None
    while not (c_ctx.complete and not s_token):
        c_token = c_ctx.step(s_token)

        if c_token:
            s_token = s_ctx.step(c_token)
        else:
            s_token = None

    assert c_ctx.complete
    assert s_ctx.complete

    return c_ctx, s_ctx



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import enum
import typing as t

from ._credential import CredHandle
from ._ntstatus import NtStatus
from ._security_buffer import SecBufferDesc

class TargetDataRep(enum.IntEnum):
    """The data representation format."""

    SECURITY_NATIVE_DREP = ...
    """The data is to be represented as the native endianess."""
    SECURITY_NETWORK_DREP = ...
    """The data is to be represented as network order or big endianess."""

class AscReq(enum.IntFlag):
    """Special context request flags.

    Special flags to request for an accept security context. These flags
    are specified in the :meth:`accept_security_context` under the
    ``context_req`` argument. The meaning and behaviour of each flag is
    controlled by the security package being used.
    """

    ASC_REQ_DELEGATE = ...
    """"The server is allowed to impersonate the client."""
    ASC_REQ_MUTUAL_AUTH = ...
    """"
    The client is required to supply a certificate to be used for client
    authentication.
    """
    ASC_REQ_REPLAY_DETECT = ...
    """Detetct replayed packets."""
    ASC_REQ_SEQUENCE_DETECT = ...
    """Detect messages received out of sequence."""
    ASC_REQ_CONFIDENTIALITY = ...
    """Encrypt and decrypt messages."""
    ASC_REQ_USE_SESSION_KEY = ...
    """A new session key must be negotiated."""
    ASC_REQ_SESSION_TICKET = ...
    ASC_REQ_ALLOCATE_MEMORY = ...
    """
    Digest and Schannel will allocate output buffers for you. When the buffers
    are deallocated they will free the memory allocated by SSPI.
    """
    ASC_REQ_USE_DCE_STYLE = ...
    """The caller expects a three-leg authentication transaction."""
    ASC_REQ_DATAGRAM = ...
    ASC_REQ_CONNECTION = ...
    """The security context will not handle formatting messages."""
    ASC_REQ_CALL_LEVEL = ...
    ASC_REQ_FRAGMENT_SUPPLIED = ...
    ASC_REQ_EXTENDED_ERROR = ...
    """When errors occur, the remote party will be notified."""
    ASC_REQ_STREAM = ...
    """Support a stream-oriented connection."""
    ASC_REQ_INTEGRITY = ...
    """Sign messages and verify signatures."""
    ASC_REQ_LICENSING = ...
    ASC_REQ_IDENTIFY = ...
    """
    When a server impersonates a context that has this flag set, the
    impersonation yields extremely limited access. Impersonation with IDENTITY
    set is used to verify the client's identity.
    """
    ASC_REQ_ALLOW_NULL_SESSION = ...
    ASC_REQ_ALLOW_NON_USER_LOGONS = ...
    ASC_REQ_ALLOW_CONTEXT_REPLAY = ...
    ASC_REQ_FRAGMENT_TO_FIT = ...
    ASC_REQ_NO_TOKEN = ...
    ASC_REQ_PROXY_BINDINGS = ...
    """
    Indicates that Digest requires channel binding. This value is mutually
    exclusive with ASC_REQ_ALLOW_MISSING_BINDINGS.
    """
    ASC_REQ_ALLOW_MISSING_BINDINGS = ...
    """
    Indicates that Digest does not require channel bindings for both inner and
    outer channels. This value is used for backward compatibility when support
    for endpoint channel binding is not known.
    """

class AscRet(enum.IntFlag):
    """Context attribute flags.

    Flags that indicate the attributes of an established context. Most flags
    correspond to the ``ASC_REQ_`` equivalents with the same name.
    """

    ASC_RET_DELEGATE = ...
    ASC_RET_MUTUAL_AUTH = ...
    ASC_RET_REPLAY_DETECT = ...
    ASC_RET_SEQUENCE_DETECT = ...
    ASC_RET_CONFIDENTIALITY = ...
    ASC_RET_USE_SESSION_KEY = ...
    ASC_RET_SESSION_TICKET = ...
    ASC_RET_ALLOCATED_MEMORY = ...
    ASC_RET_USED_DCE_STYLE = ...
    ASC_RET_DATAGRAM = ...
    ASC_RET_CONNECTION = ...
    ASC_RET_CALL_LEVEL = ...
    ASC_RET_THIRD_LEG_FAILED = ...
    ASC_RET_EXTENDED_ERROR = ...
    ASC_RET_STREAM = ...
    ASC_RET_INTEGRITY = ...
    ASC_RET_LICENSING = ...
    ASC_RET_IDENTIFY = ...
    ASC_RET_NULL_SESSION = ...
    ASC_RET_ALLOW_NON_USER_LOGONS = ...
    ASC_RET_ALLOW_CONTEXT_REPLAY = ...
    ASC_RET_FRAGMENT_ONLY = ...
    ASC_RET_NO_TOKEN = ...
    ASC_RET_NO_ADDITIONAL_TOKEN = ...

class IscReq(enum.IntFlag):
    """Special context request flags.

    Special flags to request for an initiated security context. These flags
    are specified in the :meth:`initialize_security_context` under the
    ``context_req`` argument. The meaning and behaviour of each flag is
    controlled by the security package being used.
    """

    ISC_REQ_DELEGATE = ...
    """
    The server can use the context to authenticate to other servers as the
    client. The `ISC_REQ_MUTUAL_AUTH` flag must be set for this to work. Valid
    for Kerberos. Ignore this flag for constrained delegation.
    """
    ISC_REQ_MUTUAL_AUTH = ...
    """
    The mutual authentication policy of the server will be satisified. This
    does not necessarily mean mutual authentication is performed, only that
    the authentication policy of the server is satisfied.
    """
    ISC_REQ_REPLAY_DETECT = ...
    """
    Detect replayed messages that have been encoded by using the
    :meth:`encrypt_message` or :meth:`make_signature` functions.
    """
    ISC_REQ_SEQUENCE_DETECT = ...
    """Detect messages received out of sequence."""
    ISC_REQ_CONFIDENTIALITY = ...
    """Encrypt messages by using :meth:`encrypt_message`."""
    ISC_REQ_USE_SESSION_KEY = ...
    """A new session key must be negotiated. Only supported by Kerberos."""
    ISC_REQ_PROMPT_FOR_CREDS = ...
    """
    If the client is an interactive user, the package must, if possible, prompt
    the user for the appropriate credentials"""
    ISC_REQ_USE_SUPPLIED_CREDS = ...
    """
    Schannel must not attempt to supply credentials for the client automatically.
    """
    ISC_REQ_ALLOCATE_MEMORY = ...
    """
    The security package allocated the output buffers for you. The memory is
    automatically freed when the SecBuffer is deallocated.
    """
    ISC_REQ_USE_DCE_STYLE = ...
    """The caller expects a three-leg authentication transaction."""
    ISC_REQ_DATAGRAM = ...
    """Datagram semantics must be used."""
    ISC_REQ_CONNECTION = ...
    """
    The security context will not handle formatting messages. This value is the
    default for Kerneros, Negotiate, and NTLM constrained delegation.
    """
    ISC_REQ_CALL_LEVEL = ...
    ISC_REQ_FRAGMENT_SUPPLIED = ...
    ISC_REQ_EXTENDED_ERROR = ...
    """When errors occur, the remote party will be notified."""
    ISC_REQ_STREAM = ...
    """Support a stream-oriented connection."""
    ISC_REQ_INTEGRITY = ...
    """
    Sign messages and verify signatures using the :meth:`make_signature` and
    :meth:`verify_signature` functions.
    """
    ISC_REQ_IDENTIFY = ...
    """
    When a server impersonates a context that has this flag set, that
    impersonation yields extremely limited access. Impersonation with IDENTITY
    set is used to verify the client's identity.
    """
    ISC_REQ_NULL_SESSION = ...
    ISC_REQ_MANUAL_CRED_VALIDATION = ...
    """Schannel must not authenticate the server automatically."""
    ISC_REQ_RESERVED1 = ...
    ISC_REQ_FRAGMENT_TO_FIT = ...
    ISC_REQ_FORWARD_CREDENTIALS = ...
    ISC_REQ_NO_INTEGRITY = ...
    """
    The ISC_REQ_INTEGRITY flag is ignored. This value is supported only tbe the
    Negotiate and kebreros constrained delegations.
    """
    ISC_REQ_USE_HTTP_STYLE = ...
    ISC_REQ_UNVERIFIED_TARGET_NAME = ...
    """
    The provided target name comes from an untrusted source. This controls some
    extra behaviour with Extended Protection.
    """
    ISC_REQ_CONFIDENTIALITY_ONLY = ...

class IscRet(enum.IntFlag):
    """Context attribute flags.

    Flags that indicate the attributes of an established context. Most flags
    correspond to the ``ISC_REQ_`` equivalents with the same name.
    """

    ISC_RET_DELEGATE = ...
    ISC_RET_MUTUAL_AUTH = ...
    ISC_RET_REPLAY_DETECT = ...
    ISC_RET_SEQUENCE_DETECT = ...
    ISC_RET_CONFIDENTIALITY = ...
    ISC_RET_USE_SESSION_KEY = ...
    ISC_RET_USED_COLLECTED_CREDS = ...
    ISC_RET_USED_SUPPLIED_CREDS = ...
    ISC_RET_ALLOCATED_MEMORY = ...
    ISC_RET_USED_DCE_STYLE = ...
    ISC_RET_DATAGRAM = ...
    ISC_RET_CONNECTION = ...
    ISC_RET_INTERMEDIATE_RETURN = ...
    ISC_RET_CALL_LEVEL = ...
    ISC_RET_EXTENDED_ERROR = ...
    ISC_RET_STREAM = ...
    ISC_RET_INTEGRITY = ...
    ISC_RET_IDENTIFY = ...
    ISC_RET_NULL_SESSION = ...
    ISC_RET_MANUAL_CRED_VALIDATION = ...
    ISC_RET_RESERVED1 = ...
    ISC_RET_FRAGMENT_ONLY = ...
    ISC_RET_FORWARD_CREDENTIALS = ...
    ISC_RET_USED_HTTP_STYLE = ...
    ISC_RET_NO_ADDITIONAL_TOKEN = ...
    ISC_RET_REAUTHENTICATION = ...
    ISC_RET_CONFIDENTIALITY_ONLY = ...

class CtxtHandle:
    """A security context.

    This contains the security context handle that was setup with either
    :meth:`accept_security_context` or :meth:`initialize_security_context`.
    Once no longer referenced, the handle will be freed internally, closing
    the security context.
    """

class AcceptContextResult(t.NamedTuple):
    """The accept security context result."""

    context: CtxtHandle
    """
    The generated context to use for subsequent operations on this context.
    This context should be passed as the `context` arg on any remaining calls
    to :meth:`accept_security_context`.
    """
    attributes: AscRet
    """
    A set of flags that indicate the attributes of the established context.
    Security-related attributes should only be checked until the status
    ``SEC_E_OK`` has been received. Non-security related attributes can be
    checked before the context is fully established.
    """
    expiry: int
    """The time at which the context expires as a FILETIME value."""
    status: NtStatus
    """The NtStatus code result of the operation."""

class InitializeContextResult(t.NamedTuple):
    """The initialize security context result."""

    context: CtxtHandle
    """
    The generated context to use for subsequent operations on this context.
    This context should be passed as the `context` arg on any remaining calls
    to :meth:`initialize_security_context`.
    """
    attributes: IscRet
    """
    A set of flags that indicate the attributes of the established context.
    Security-related attributes should only be checked until the status
    ``SEC_E_OK`` has been received. Non-security related attributes can be
    checked before the context is fully established.
    """
    expiry: int
    """The time at which the context expires as a FILETIME value."""
    status: NtStatus
    """The NtStatus code result of the operation."""

def accept_security_context(
    credential: CredHandle | None,
    context: CtxtHandle | None,
    input_buffers: SecBufferDesc | None,
    context_req: AscReq | int,
    target_data_rep: TargetDataRep | int,
    output_buffers: SecBufferDesc | None,
) -> AcceptContextResult:
    """Initiates the server side security context.

    This function initiates the server side, inbound security context from a
    credentials handle. It is used to build a security context between the
    application and a remote peer.

    The function should be called in a loop until a sufficient security
    context is established. The security tokens generated in the output_buffers
    should be exchanged with the peer and any tokens received from that peer
    should be placed in the input_buffers. The values that can be returned as
    the result values are:

        SEC_E_OK - the context was initialized
        SEC_I_CONTINUE_NEEDED - more input tokens are needed
        SEC_I_COMPLETE_NEEDED - context needs to be completed
        SEC_I_COMPLETE_AND_CONTINUE - more tokens and must be completed

    To complete a security context call :meth:`complete_auth_token`. Currently
    only the Digest security provider is known to require completion.

    The credential can be a credential generated by
    :meth:`acquire_credentials_handle` or ``None``. If ``None``, the current
    user context will be used. A credential can control what security provider
    is used for authentication so should be set in most circumstances.

    The context should be set to ``None`` on the first call and the returned
    context result on subsequent calls.

    The input buffers contains any input tokens, or other buffer types
    supported by the security package. It is typically set to the data received
    by the client. It may contain more buffer types like
    ``SECBUFFER_CHANNEL_BINDINGS``.

    The context req are a set of :class:`AscReq` flags for the context.
    The behaviour of these flags are dependent on the security provider being
    used.

    The target_data_rep specifies the data representation. Typically this is
    just set to ``SECURITY_NETWORK_DREP``.

    The output buffers contain the buffers which will store the output tokens
    generated by the security package. The buffers must be large enough to
    contain the data that could be generated by that buffer type or the
    context_req ``ASC_REQ_ALLOCATE_MEMORY`` is specified.

    This function wraps the `AcceptSecurityContext`_ Win32 function.

    Args:
        credential: Optional credential to use for this security exchange.
        context: Must be None on the first call and the returned security
            context on the next call(s)
        input_buffers: The buffers supplied as input to the package.
        context_req: Specify special request flags for the context.
        target_data_rep: Specifies the data representation, such as byte
            ordering, on the target.
        output_buffers: The buffers that will contain the output from the
            security package.

    Returns:
        AcceptorContextResult: Contains the security context, attributes,
        expiry, and NtStatus result from the call.

    Raises:
        WindowsError: If the function failed.

    .. _AcceptSecurityContext:
        https://learn.microsoft.com/en-us/windows/win32/secauthn/acceptsecuritycontext--general
    """

def complete_auth_token(
    context: CtxtHandle,
    token: SecBufferDesc,
) -> None:
    """Completes an authentication token.

    This function is used by protocols, such as DCE, that need to revise the
    security information after the transport application has updated some
    message parameters.

    This function is supported only by the Digest security provider and is used
    on the server side only.

    This function wraps the `CompleteAuthToken`_ Win32 function.

    Args:
        context: A handle of the context that needs to be completed.
        token: A SecBufferDesc that contains the token to be completed.

    Raises:
        WindowsError: If the function failed.

    .. _CompleteAuthToken:
        https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-completeauthtoken
    """

def initialize_security_context(
    credential: CredHandle | None,
    context: CtxtHandle | None,
    target_name: str,
    context_req: IscReq | int,
    target_data_rep: TargetDataRep | int,
    input_buffers: SecBufferDesc | None,
    output_buffers: SecBufferDesc | None,
) -> InitializeContextResult:
    """Initiates the client side security context.

    This function initiates the client side, outbound security context from a
    credentials handle. It is used to build a security context between the
    application and a remote peer.

    The function should be called in a loop until a sufficient security
    context is established. The security tokens generated in the output_buffers
    should be exchanged with the peer and any tokens received from that peer
    should be placed in the input_buffers. The values that can be returned as
    the result values are:

        SEC_E_OK - the context was initialized
        SEC_I_CONTINUE_NEEDED - more input tokens are needed
        SEC_I_COMPLETE_NEEDED - context needs to be completed
        SEC_I_COMPLETE_AND_CONTINUE - more tokens and must be completed

    To complete a security context call :meth:`complete_auth_token`. Currently
    only the Digest security provider is known to require completion.

    The credential can be a credential generated by
    :meth:`acquire_credentials_handle` or ``None``. If ``None``, the current
    user context will be used. A credential can control what security provider
    is used for authentication so should be set in most circumstances.

    The context should be set to ``None`` on the first call and the returned
    context result on subsequent calls.

    The target name is set to a security provider specific format. For example
    with Kerberos, Negotiate, and NTLM it should be set to the Service
    Principal Name (SPN) or the target. For Schannel/TLS it should be set to
    the target hostname used for certificate name verification.

    The context req are a set of :class:`IscReq` flags for the context.
    The behaviour of these flags are dependent on the security provider being
    used.

    The target_data_rep specifies the data representation. Typically this is
    just set to ``SECURITY_NETWORK_DREP``.

    The input buffers contains any input tokens, or other buffer types
    supported by the security package. It is typically set to ``None` on the
    first call and contains the server response token on subsequent calls.
    Other buffers like ``SECBUFFER_CHANNEL_BINDINGS`` can be specified here
    dependent on the security provider used.

    The output buffers contain the buffers which will store the output tokens
    generated by the security package. The buffers must be large enough to
    contain the data that could be generated by that buffer type or the
    context_req ``ISC_REQ_ALLOCATE_MEMORY`` is specified.

    This function wraps the `InitializeSecurityContext`_ Win32 function.

    Args:
        credential: Optional credential to use for this security exchange.
        context: Must be None on the first call and the returned security
            context on the next call(s)
        target_name: The target of the context.
        context_req: Specify special request flags for the context.
        target_data_rep: Specifies the data representation, such as byte
            ordering, on the target.
        input_buffers: The buffers supplied as input to the package.
        output_buffers: The buffers that will contain the output from the
            security package.

    Returns:
        InitializeContextResult: Contains the security context, attributes,
        expiry, and NtStatus result from the call.

    Raises:
        WindowsError: If the function failed.

    .. _InitializeSecurityContext:
        https://learn.microsoft.com/en-us/windows/win32/secauthn/initializesecuritycontext--general
    """



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

from ._win32_types cimport *


cdef extern from "python_sspi.h":
    cdef struct _SecPkgInfoW:
        unsigned int fCapabilities
        unsigned short wVersion
        unsigned short wRPCID
        unsigned int cbMaxToken
        SEC_WCHAR *Name
        SEC_WCHAR *Comment
    ctypedef _SecPkgInfoW SecPkgInfoW
    ctypedef SecPkgInfoW *PSecPkgInfoW



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import os

import sspi.raw as sr


def test_set_kdc_proxy_default() -> None:
    kdc_proxy = sr.SecPkgCredKdcProxySettings()
    assert kdc_proxy.version == 1
    assert kdc_proxy.flags == sr.KdcProxySettingsFlags.KDX_PROXY_SETTINGS_FLAGS_NONE
    assert kdc_proxy.proxy_server is None
    assert repr(kdc_proxy) == "SecPkgCredKdcProxySettings(flags=0, proxy_server=None)"

    # While we can't verify it we can at least make sure it doesn't fail to set.
    auth_data = None
    if os.name != "nt":
        auth_data = sr.WinNTAuthIdentity(username="user", password="pass")
    cred = sr.acquire_credentials_handle(
        None,
        "Kerberos",
        sr.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_data,
    )[0]
    sr.set_credentials_attributes(cred, kdc_proxy)


def test_set_kdc_proxy_flags() -> None:
    kdc_proxy = sr.SecPkgCredKdcProxySettings(flags=sr.KdcProxySettingsFlags.KDC_PROXY_SETTINGS_FLAGS_FORCEPROXY)
    assert kdc_proxy.version == 1
    assert kdc_proxy.flags == sr.KdcProxySettingsFlags.KDC_PROXY_SETTINGS_FLAGS_FORCEPROXY
    assert kdc_proxy.proxy_server is None
    assert repr(kdc_proxy) == "SecPkgCredKdcProxySettings(flags=1, proxy_server=None)"

    # While we can't verify it we can at least make sure it doesn't fail to set.
    auth_data = None
    if os.name != "nt":
        auth_data = sr.WinNTAuthIdentity(username="user", password="pass")
    cred = sr.acquire_credentials_handle(
        None,
        "Kerberos",
        sr.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_data,
    )[0]
    sr.set_credentials_attributes(cred, kdc_proxy)


def test_set_kdc_proxy_server() -> None:
    # Test the length calcs work with NULL and surrogate pairs
    kdc_proxy = sr.SecPkgCredKdcProxySettings(proxy_server="kdc.\U0001F4A0.com:443:kdc\u0000proxy")
    assert kdc_proxy.version == 1
    assert kdc_proxy.flags == sr.KdcProxySettingsFlags.KDX_PROXY_SETTINGS_FLAGS_NONE
    assert kdc_proxy.proxy_server == "kdc.\U0001F4A0.com:443:kdc\u0000proxy"
    assert repr(kdc_proxy) == "SecPkgCredKdcProxySettings(flags=0, proxy_server='kdc.\U0001F4A0.com:443:kdc\\x00proxy')"

    # While we can't verify it we can at least make sure it doesn't fail to set.
    auth_data = None
    if os.name != "nt":
        auth_data = sr.WinNTAuthIdentity(username="user", password="pass")
    cred = sr.acquire_credentials_handle(
        None,
        "Kerberos",
        sr.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_data,
    )[0]
    sr.set_credentials_attributes(cred, kdc_proxy)



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import sspi


def test_cred_no_user() -> None:
    cred = sspi.UserCredential()
    assert str(cred) == "CurrentUser - Negotiate"
    assert cred.username == "CurrentUser"


def test_cred_username_no_domain() -> None:
    cred = sspi.UserCredential("user")
    assert str(cred) == "user - Negotiate"
    assert cred.username == "user"


def test_cred_username_upn() -> None:
    cred = sspi.UserCredential("user@DOMAIN.COM")
    assert str(cred) == "user@DOMAIN.COM - Negotiate"
    assert cred.username == "user@DOMAIN.COM"


def test_cred_username_domain() -> None:
    cred = sspi.UserCredential("user", domain="domain")
    assert str(cred) == "domain\\user - Negotiate"
    assert cred.username == "domain\\user"



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import sspi.raw as raw
from sspi.raw import AscReq, AscRet, IscReq, IscRet, SecChannelBindings, WindowsError

from ._credential import UserCredential
from ._sec_context import (
    ClientSecurityContext,
    SecurityContext,
    ServerSecurityContext,
    UnwrapResult,
)

__all__ = [
    "AscReq",
    "AscRet",
    "IscReq",
    "IscRet",
    "ClientSecurityContext",
    "SecChannelBindings",
    "SecurityContext",
    "ServerSecurityContext",
    "UnwrapResult",
    "UserCredential",
    "WindowsError",
    "raw",
]



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import os

import pytest

import sspi.raw as sr


@pytest.mark.skipif(os.name != "nt", reason="sspi-rs does not support signature functions")
def test_sign_and_verify(
    authenticated_contexts: tuple[sr.CtxtHandle, sr.CtxtHandle],
) -> None:
    client_data = os.urandom(32)
    server_data = os.urandom(32)

    sizes = sr.query_context_attributes(authenticated_contexts[0], sr.SecPkgContextSizes)

    in_data = bytearray(client_data)
    in_token = bytearray(sizes.security_trailer)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
            sr.SecBuffer(in_token, sr.SecBufferType.SECBUFFER_TOKEN),
        ],
    )
    sr.make_signature(
        authenticated_contexts[0],
        0,
        in_message,
        0,
    )

    assert bytes(in_data) == client_data
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_data) == in_message[0].data

    assert bytes(in_token) != b"\x00" * sizes.security_trailer
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_token) == in_message[1].data

    res = sr.verify_signature(
        authenticated_contexts[1],
        in_message,
        0,
    )
    assert res == 0

    assert bytes(in_data) == client_data
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_data) == in_message[0].data

    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_token) == in_message[1].data

    in_token = bytearray(sizes.security_trailer)
    in_data = bytearray(server_data)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
            sr.SecBuffer(in_token, sr.SecBufferType.SECBUFFER_TOKEN),
        ],
    )
    sr.make_signature(
        authenticated_contexts[1],
        sr.QopFlags(0),
        in_message,
        0,
    )

    assert bytes(in_data) == server_data
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_data) == in_message[0].data

    assert bytes(in_token) != b"\x00" * sizes.security_trailer
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_token) == in_message[1].data

    res = sr.verify_signature(
        authenticated_contexts[0],
        in_message,
        0,
    )
    assert res == 0

    assert bytes(in_data) == server_data
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_data) == in_message[0].data

    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_token) == in_message[1].data


@pytest.mark.skipif(os.name != "nt", reason="sspi-rs does not support signature functions")
def test_make_signature_fail(
    authenticated_contexts: tuple[sr.CtxtHandle, sr.CtxtHandle],
) -> None:
    data = b"message"

    in_data = bytearray(data)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
        ],
    )

    with pytest.raises(sr.WindowsError) as e:
        sr.make_signature(
            authenticated_contexts[0],
            0,
            in_message,
            0,
        )

    assert e.value.winerror == -2146893048  # SEC_E_INVALID_TOKEN


@pytest.mark.skipif(os.name != "nt", reason="sspi-rs does not support signature functions")
def test_verify_failure(
    authenticated_contexts: tuple[sr.CtxtHandle, sr.CtxtHandle],
) -> None:
    data = b"message"

    sizes = sr.query_context_attributes(authenticated_contexts[0], sr.SecPkgContextSizes)

    in_data = bytearray(data)
    in_token = bytearray(sizes.security_trailer)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
            sr.SecBuffer(in_token, sr.SecBufferType.SECBUFFER_TOKEN),
        ],
    )
    sr.make_signature(
        authenticated_contexts[0],
        0,
        in_message,
        0,
    )

    in_data[0] = 0

    with pytest.raises(sr.WindowsError) as e:
        sr.verify_signature(
            authenticated_contexts[1],
            in_message,
            0,
        )

    assert e.value.winerror == -2146893041  # SEC_E_MESSAGE_ALTERED


def test_encrypt_and_decrypt(
    authenticated_contexts: tuple[sr.CtxtHandle, sr.CtxtHandle],
) -> None:
    client_data = os.urandom(32)
    server_data = os.urandom(32)

    sizes = sr.query_context_attributes(authenticated_contexts[0], sr.SecPkgContextSizes)

    in_token = bytearray(sizes.security_trailer)
    in_data = bytearray(client_data)
    in_padding = bytearray(sizes.block_size)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_token, sr.SecBufferType.SECBUFFER_TOKEN),
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
            sr.SecBuffer(in_padding, sr.SecBufferType.SECBUFFER_PADDING),
        ],
    )
    sr.encrypt_message(
        authenticated_contexts[0],
        sr.QopFlags(0),
        in_message,
        0,
    )

    assert bytes(in_token) != b"\x00" * sizes.security_trailer
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_token) == in_message[0].data

    assert bytes(in_data) != client_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data

    res = sr.decrypt_message(
        authenticated_contexts[1],
        in_message,
        0,
    )
    assert res == 0

    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_token) == in_message[0].data

    assert bytes(in_data) == client_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data

    in_token = bytearray(sizes.security_trailer)
    in_data = bytearray(server_data)
    in_padding = bytearray(sizes.block_size)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_token, sr.SecBufferType.SECBUFFER_TOKEN),
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
            sr.SecBuffer(in_padding, sr.SecBufferType.SECBUFFER_PADDING),
        ],
    )
    sr.encrypt_message(
        authenticated_contexts[1],
        sr.QopFlags(0),
        in_message,
        0,
    )

    assert bytes(in_token) != b"\x00" * sizes.security_trailer
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_token) == in_message[0].data

    assert bytes(in_data) != server_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data

    res = sr.decrypt_message(
        authenticated_contexts[0],
        in_message,
        0,
    )
    assert res == 0

    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_token) == in_message[0].data

    assert bytes(in_data) == server_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data


def test_encrypt_message_fail(
    authenticated_contexts: tuple[sr.CtxtHandle, sr.CtxtHandle],
) -> None:
    data = b"message"

    in_data = bytearray(data)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
        ],
    )

    with pytest.raises(sr.WindowsError) as e:
        sr.encrypt_message(
            authenticated_contexts[0],
            0,
            in_message,
            0,
        )

    assert e.value.winerror == -2146893048  # SEC_E_INVALID_TOKEN


def test_decrypt_message_failure(
    authenticated_contexts: tuple[sr.CtxtHandle, sr.CtxtHandle],
) -> None:
    data = b"message"

    sizes = sr.query_context_attributes(authenticated_contexts[0], sr.SecPkgContextSizes)

    in_data = bytearray(data)
    in_token = bytearray(sizes.security_trailer)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
            sr.SecBuffer(in_token, sr.SecBufferType.SECBUFFER_TOKEN),
        ],
    )
    sr.encrypt_message(
        authenticated_contexts[0],
        0,
        in_message,
        0,
    )

    in_data[0] = 0 if in_data[0] == 255 else in_data[0] + 1

    with pytest.raises(sr.WindowsError) as e:
        sr.decrypt_message(
            authenticated_contexts[1],
            in_message,
            0,
        )

    assert e.value.winerror == -2146893041  # SEC_E_MESSAGE_ALTERED


# https://github.com/Devolutions/sspi-rs/issues/84
@pytest.mark.skipif(os.name != "nt", reason="SECBUFFER_STREAM not support with NTLM in sspi-rs")
def test_encrypt_and_decrypt_stream(
    authenticated_contexts: tuple[sr.CtxtHandle, sr.CtxtHandle],
) -> None:
    client_data = os.urandom(32)
    server_data = os.urandom(32)

    sizes = sr.query_context_attributes(authenticated_contexts[0], sr.SecPkgContextSizes)

    in_token = bytearray(sizes.security_trailer)
    in_data = bytearray(client_data)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_token, sr.SecBufferType.SECBUFFER_TOKEN),
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
        ],
    )
    sr.encrypt_message(
        authenticated_contexts[0],
        sr.QopFlags(0),
        in_message,
        0,
    )

    assert bytes(in_token) != b"\x00" * sizes.security_trailer
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_token) == in_message[0].data

    assert bytes(in_data) != client_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data

    stream_data = bytearray(
        b"".join(
            [
                in_message[0].data,
                in_message[1].data,
            ]
        )
    )
    stream_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(stream_data, sr.SecBufferType.SECBUFFER_STREAM),
            sr.SecBuffer(None, sr.SecBufferType.SECBUFFER_DATA),
        ]
    )
    res = sr.decrypt_message(
        authenticated_contexts[1],
        stream_message,
        0,
    )
    assert res == 0

    assert stream_message[1].data == client_data

    in_token = bytearray(sizes.security_trailer)
    in_data = bytearray(server_data)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_token, sr.SecBufferType.SECBUFFER_TOKEN),
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
        ],
    )
    sr.encrypt_message(
        authenticated_contexts[1],
        sr.QopFlags(0),
        in_message,
        0,
    )

    assert bytes(in_token) != b"\x00" * sizes.security_trailer
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_token) == in_message[0].data

    assert bytes(in_data) != server_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data

    stream_data = bytearray(
        b"".join(
            [
                in_message[0].data,
                in_message[1].data,
            ]
        )
    )
    stream_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(stream_data, sr.SecBufferType.SECBUFFER_STREAM),
            sr.SecBuffer(None, sr.SecBufferType.SECBUFFER_DATA),
        ]
    )
    res = sr.decrypt_message(
        authenticated_contexts[0],
        stream_message,
        0,
    )
    assert res == 0

    assert stream_message[1].data == server_data


def test_encrypt_winrm(
    authenticated_contexts: tuple[sr.CtxtHandle, sr.CtxtHandle],
) -> None:
    client_data = os.urandom(32)
    server_data = os.urandom(32)

    sizes = sr.query_context_attributes(authenticated_contexts[0], sr.SecPkgContextSizes)

    in_token = bytearray(sizes.security_trailer)
    in_data = bytearray(client_data)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_token, sr.SecBufferType.SECBUFFER_TOKEN),
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
        ],
    )
    sr.encrypt_message(
        authenticated_contexts[0],
        sr.QopFlags(0),
        in_message,
        0,
    )

    assert bytes(in_token) != b"\x00" * sizes.security_trailer
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_token) == in_message[0].data

    assert bytes(in_data) != client_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data

    res = sr.decrypt_message(
        authenticated_contexts[1],
        in_message,
        0,
    )
    assert res == 0

    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_token) == in_message[0].data

    assert bytes(in_data) == client_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data

    in_token = bytearray(sizes.security_trailer)
    in_data = bytearray(server_data)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_token, sr.SecBufferType.SECBUFFER_TOKEN),
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
        ],
    )
    sr.encrypt_message(
        authenticated_contexts[1],
        sr.QopFlags(0),
        in_message,
        0,
    )

    assert bytes(in_token) != b"\x00" * sizes.security_trailer
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_token) == in_message[0].data

    assert bytes(in_data) != server_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data

    res = sr.decrypt_message(
        authenticated_contexts[0],
        in_message,
        0,
    )
    assert res == 0

    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_token) == in_message[0].data

    assert bytes(in_data) == server_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data


# https://github.com/Devolutions/sspi-rs/issues/120
@pytest.mark.skipif(os.name != "nt", reason="No READONLY buffer support in sspi-rs")
@pytest.mark.parametrize(
    ["sign_header"],
    [
        (True,),
        (False,),
    ],
)
def test_encrypt_dce(
    sign_header: bool,
    authenticated_contexts: tuple[sr.CtxtHandle, sr.CtxtHandle],
) -> None:
    buffer_flags = (
        sr.SecBufferFlags.SECBUFFER_READONLY_WITH_CHECKSUM if sign_header else sr.SecBufferFlags.SECBUFFER_READONLY
    )
    client_data = os.urandom(32)
    client_header = os.urandom(24)
    client_trailer = os.urandom(8)

    server_data = os.urandom(32)
    server_header = os.urandom(24)
    server_trailer = os.urandom(8)

    sizes = sr.query_context_attributes(authenticated_contexts[0], sr.SecPkgContextSizes)

    in_header = bytearray(client_header)
    in_data = bytearray(client_data)
    in_trailer = bytearray(client_trailer)
    in_token = bytearray(sizes.security_trailer)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_header, sr.SecBufferType.SECBUFFER_DATA, buffer_flags),
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
            sr.SecBuffer(in_trailer, sr.SecBufferType.SECBUFFER_DATA, buffer_flags),
            sr.SecBuffer(in_token, sr.SecBufferType.SECBUFFER_TOKEN),
        ],
    )
    sr.encrypt_message(
        authenticated_contexts[0],
        sr.QopFlags(0),
        in_message,
        0,
    )

    assert bytes(in_header) == client_header
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_header) == in_message[0].data

    assert bytes(in_data) != client_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data

    assert bytes(in_trailer) == client_trailer
    assert len(in_message[2].data) == in_message[2].count
    assert bytes(in_trailer) == in_message[2].data

    assert bytes(in_token) != b"\x00" * sizes.security_trailer
    assert len(in_message[3].data) == in_message[3].count
    assert bytes(in_token) == in_message[3].data

    res = sr.decrypt_message(
        authenticated_contexts[1],
        in_message,
        0,
    )
    assert res == 0

    assert bytes(in_header) == client_header
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_header) == in_message[0].data

    assert bytes(in_data) == client_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data

    assert bytes(in_trailer) == client_trailer
    assert len(in_message[2].data) == in_message[2].count
    assert bytes(in_trailer) == in_message[2].data

    assert len(in_message[3].data) == in_message[3].count
    assert bytes(in_token) == in_message[3].data

    in_header = bytearray(server_header)
    in_data = bytearray(server_data)
    in_trailer = bytearray(server_trailer)
    in_token = bytearray(sizes.security_trailer)
    in_message = sr.SecBufferDesc(
        [
            sr.SecBuffer(in_header, sr.SecBufferType.SECBUFFER_DATA, buffer_flags),
            sr.SecBuffer(in_data, sr.SecBufferType.SECBUFFER_DATA),
            sr.SecBuffer(in_trailer, sr.SecBufferType.SECBUFFER_DATA, buffer_flags),
            sr.SecBuffer(in_token, sr.SecBufferType.SECBUFFER_TOKEN),
        ],
    )
    sr.encrypt_message(
        authenticated_contexts[1],
        sr.QopFlags(0),
        in_message,
        0,
    )

    assert bytes(in_header) == server_header
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_header) == in_message[0].data

    assert bytes(in_data) != server_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data

    assert bytes(in_trailer) == server_trailer
    assert len(in_message[2].data) == in_message[2].count
    assert bytes(in_trailer) == in_message[2].data

    assert bytes(in_token) != b"\x00" * sizes.security_trailer
    assert len(in_message[3].data) == in_message[3].count
    assert bytes(in_token) == in_message[3].data

    res = sr.decrypt_message(
        authenticated_contexts[0],
        in_message,
        0,
    )
    assert res == 0

    assert bytes(in_header) == server_header
    assert len(in_message[0].data) == in_message[0].count
    assert bytes(in_header) == in_message[0].data

    assert bytes(in_data) == server_data
    assert len(in_message[1].data) == in_message[1].count
    assert bytes(in_data) == in_message[1].data

    assert bytes(in_trailer) == server_trailer
    assert len(in_message[2].data) == in_message[2].count
    assert bytes(in_trailer) == in_message[2].data

    assert len(in_message[3].data) == in_message[3].count
    assert bytes(in_token) == in_message[3].data



# Copyright: (c) 2023, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)


cdef extern from "python_sspi.h":
    unsigned int _SEC_I_COMPLETE_AND_CONTINUE "SEC_I_COMPLETE_AND_CONTINUE"
    unsigned int _SEC_I_COMPLETE_NEEDED "SEC_I_COMPLETE_NEEDED"
    unsigned int _SEC_I_CONTINUE_NEEDED "SEC_I_CONTINUE_NEEDED"
    unsigned int _SEC_E_OK "SEC_E_OK"



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import os
import socket

import pytest

import sspi.raw as sr


def test_sec_exchange_alloc_memory() -> None:
    spn = f"host/{socket.gethostname()}"

    auth_data = None
    server_package = "Negotiate"
    if os.name != "nt":
        # sspi-rs needs NTLM on the server with explicit creds
        auth_data = sr.WinNTAuthIdentity(
            username="user",
            password="pass",
        )
        server_package = "NTLM"

    c_cred = sr.acquire_credentials_handle(
        None,
        "NTLM",
        sr.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_data,
    )[0]
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(None, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sr.initialize_security_context(
        credential=c_cred,
        context=None,
        target_name=spn,
        context_req=sr.IscReq.ISC_REQ_ALLOCATE_MEMORY,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=None,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sr.InitializeContextResult)
    assert isinstance(c_res.context, sr.CtxtHandle)
    assert isinstance(c_res.attributes, sr.IscRet)
    assert isinstance(c_res.expiry, int)
    assert isinstance(c_res.status, sr.NtStatus)
    assert c_res.status == sr.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    s_cred = sr.acquire_credentials_handle(
        None,
        server_package,
        sr.CredentialUse.SECPKG_CRED_INBOUND,
        auth_data=auth_data,
    )[0]
    s_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(c_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(None, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sr.accept_security_context(
        credential=s_cred,
        context=None,
        input_buffers=s_input_buffers,
        context_req=sr.AscReq.ASC_REQ_ALLOCATE_MEMORY,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sr.AcceptContextResult)
    assert isinstance(s_res.context, sr.CtxtHandle)
    assert isinstance(s_res.attributes, sr.AscRet)
    assert isinstance(s_res.expiry, int)
    assert isinstance(s_res.status, sr.NtStatus)
    assert s_res.status == sr.NtStatus.SEC_I_CONTINUE_NEEDED

    assert s_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert s_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    c_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(s_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(None, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sr.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=sr.IscReq.ISC_REQ_ALLOCATE_MEMORY,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sr.InitializeContextResult)
    assert isinstance(c_res.context, sr.CtxtHandle)
    assert isinstance(c_res.attributes, sr.IscRet)
    assert isinstance(c_res.expiry, int)
    assert isinstance(c_res.status, sr.NtStatus)
    assert c_res.status == sr.NtStatus.SEC_E_OK

    assert c_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    s_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(c_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sr.accept_security_context(
        credential=s_cred,
        context=s_res.context,
        input_buffers=s_input_buffers,
        context_req=sr.AscReq.ASC_REQ_ALLOCATE_MEMORY,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sr.AcceptContextResult)
    assert isinstance(s_res.context, sr.CtxtHandle)
    assert isinstance(s_res.attributes, sr.AscRet)
    assert isinstance(s_res.expiry, int)
    assert isinstance(s_res.status, sr.NtStatus)

    if os.name == "nt":
        assert s_res.status == sr.NtStatus.SEC_E_OK
    else:
        # https://github.com/Devolutions/sspi-rs/issues/167
        assert s_res.status == sr.NtStatus.SEC_I_COMPLETE_NEEDED
        sr.complete_auth_token(s_res.context, s_input_buffers)


def test_sec_exchange_prealloc_mem() -> None:
    spn = f"host/{socket.gethostname()}"
    buffer = bytearray(4096)

    auth_data = None
    server_package = "Negotiate"
    if os.name != "nt":
        # sspi-rs needs NTLM on the server with explicit creds
        auth_data = sr.WinNTAuthIdentity(
            username="user",
            password="pass",
        )
        server_package = "NTLM"

    c_cred = sr.acquire_credentials_handle(
        None,
        "NTLM",
        sr.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_data,
    )[0]
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sr.initialize_security_context(
        credential=c_cred,
        context=None,
        target_name=spn,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=None,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sr.InitializeContextResult)
    assert isinstance(c_res.context, sr.CtxtHandle)
    assert isinstance(c_res.attributes, sr.IscRet)
    assert isinstance(c_res.expiry, int)
    assert isinstance(c_res.status, sr.NtStatus)
    assert c_res.status == sr.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    s_cred = sr.acquire_credentials_handle(
        None,
        server_package,
        sr.CredentialUse.SECPKG_CRED_INBOUND,
        auth_data=auth_data,
    )[0]
    s_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(c_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sr.accept_security_context(
        credential=s_cred,
        context=None,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sr.AcceptContextResult)
    assert isinstance(s_res.context, sr.CtxtHandle)
    assert isinstance(s_res.attributes, sr.AscRet)
    assert isinstance(s_res.expiry, int)
    assert isinstance(s_res.status, sr.NtStatus)
    assert s_res.status == sr.NtStatus.SEC_I_CONTINUE_NEEDED

    assert s_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert s_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    c_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(s_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sr.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sr.InitializeContextResult)
    assert isinstance(c_res.context, sr.CtxtHandle)
    assert isinstance(c_res.attributes, sr.IscRet)
    assert isinstance(c_res.expiry, int)
    assert isinstance(c_res.status, sr.NtStatus)
    assert c_res.status == sr.NtStatus.SEC_E_OK

    assert c_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    s_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(c_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sr.accept_security_context(
        credential=s_cred,
        context=s_res.context,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sr.AcceptContextResult)
    assert isinstance(s_res.context, sr.CtxtHandle)
    assert isinstance(s_res.attributes, sr.AscRet)
    assert isinstance(s_res.expiry, int)
    assert isinstance(s_res.status, sr.NtStatus)

    if os.name == "nt":
        assert s_res.status == sr.NtStatus.SEC_E_OK
    else:
        # https://github.com/Devolutions/sspi-rs/issues/167
        assert s_res.status == sr.NtStatus.SEC_I_COMPLETE_NEEDED
        sr.complete_auth_token(s_res.context, s_input_buffers)


# https://github.com/Devolutions/sspi-rs/issues/172
@pytest.mark.skipif(os.name != "nt", reason="Seems like sspi-rs doesn't validate the credentials")
def test_invalid_credential() -> None:
    spn = f"host/{socket.gethostname()}"
    buffer = bytearray(4096)

    invalid_cred = sr.WinNTAuthIdentity(username="invalid", password="invalid")
    c_cred = sr.acquire_credentials_handle(
        None,
        "NTLM",
        sr.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=invalid_cred,
    )[0]
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sr.initialize_security_context(
        credential=c_cred,
        context=None,
        target_name=spn,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=None,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sr.InitializeContextResult)
    assert isinstance(c_res.context, sr.CtxtHandle)
    assert isinstance(c_res.attributes, sr.IscRet)
    assert isinstance(c_res.expiry, int)
    assert isinstance(c_res.status, sr.NtStatus)
    assert c_res.status == sr.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    if os.name == "nt":
        s_cred = sr.acquire_credentials_handle(
            None,
            "Negotiate",
            sr.CredentialUse.SECPKG_CRED_INBOUND,
        )[0]
    else:
        s_cred = sr.acquire_credentials_handle(
            None,
            "NTLM",
            sr.CredentialUse.SECPKG_CRED_INBOUND,
            auth_data=sr.WinNTAuthIdentity(username="invalid", password="other"),
        )[0]
    s_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(c_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sr.accept_security_context(
        credential=s_cred,
        context=None,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sr.AcceptContextResult)
    assert isinstance(s_res.context, sr.CtxtHandle)
    assert isinstance(s_res.attributes, sr.AscRet)
    assert isinstance(s_res.expiry, int)
    assert isinstance(s_res.status, sr.NtStatus)
    assert s_res.status == sr.NtStatus.SEC_I_CONTINUE_NEEDED

    assert s_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert s_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    c_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(s_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sr.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sr.InitializeContextResult)
    assert isinstance(c_res.context, sr.CtxtHandle)
    assert isinstance(c_res.attributes, sr.IscRet)
    assert isinstance(c_res.expiry, int)
    assert isinstance(c_res.status, sr.NtStatus)
    assert c_res.status == sr.NtStatus.SEC_E_OK

    assert c_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    s_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(c_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    with pytest.raises(sr.WindowsError) as e:
        sr.accept_security_context(
            credential=s_cred,
            context=s_res.context,
            input_buffers=s_input_buffers,
            context_req=0,
            target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
            output_buffers=s_output_buffers,
        )

    assert e.value.winerror == -2146893044  # SEC_E_LOGON_DENIED


def test_sec_exchange_with_channel_binding() -> None:
    spn = f"host/{socket.gethostname()}"
    buffer = bytearray(4096)
    channel_bindings = sr.SecChannelBindings(
        application_data=b"tls-server-end-point:" + (b"\x00" * 32),
    ).get_sec_buffer_copy()

    auth_data = None
    server_package = "Negotiate"
    if os.name != "nt":
        # sspi-rs needs NTLM on the server with explicit creds
        auth_data = sr.WinNTAuthIdentity(
            username="user",
            password="pass",
        )
        server_package = "NTLM"

    c_cred = sr.acquire_credentials_handle(
        None,
        "NTLM",
        sr.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_data,
    )[0]
    c_input_buffers = sr.SecBufferDesc([channel_bindings])
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sr.initialize_security_context(
        credential=c_cred,
        context=None,
        target_name=spn,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sr.InitializeContextResult)
    assert isinstance(c_res.context, sr.CtxtHandle)
    assert isinstance(c_res.attributes, sr.IscRet)
    assert isinstance(c_res.expiry, int)
    assert isinstance(c_res.status, sr.NtStatus)
    assert c_res.status == sr.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    s_cred = sr.acquire_credentials_handle(
        None,
        server_package,
        sr.CredentialUse.SECPKG_CRED_INBOUND,
        auth_data=auth_data,
    )[0]
    s_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(c_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
            channel_bindings,
        ]
    )
    s_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sr.accept_security_context(
        credential=s_cred,
        context=None,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sr.AcceptContextResult)
    assert isinstance(s_res.context, sr.CtxtHandle)
    assert isinstance(s_res.attributes, sr.AscRet)
    assert isinstance(s_res.expiry, int)
    assert isinstance(s_res.status, sr.NtStatus)
    assert s_res.status == sr.NtStatus.SEC_I_CONTINUE_NEEDED

    assert s_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert s_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    c_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(s_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
            channel_bindings,
        ]
    )
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sr.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sr.InitializeContextResult)
    assert isinstance(c_res.context, sr.CtxtHandle)
    assert isinstance(c_res.attributes, sr.IscRet)
    assert isinstance(c_res.expiry, int)
    assert isinstance(c_res.status, sr.NtStatus)
    assert c_res.status == sr.NtStatus.SEC_E_OK

    assert c_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    s_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(c_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
            channel_bindings,
        ]
    )
    s_res = sr.accept_security_context(
        credential=s_cred,
        context=s_res.context,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sr.AcceptContextResult)
    assert isinstance(s_res.context, sr.CtxtHandle)
    assert isinstance(s_res.attributes, sr.AscRet)
    assert isinstance(s_res.expiry, int)
    assert isinstance(s_res.status, sr.NtStatus)

    if os.name == "nt":
        assert s_res.status == sr.NtStatus.SEC_E_OK
    else:
        # https://github.com/Devolutions/sspi-rs/issues/167
        assert s_res.status == sr.NtStatus.SEC_I_COMPLETE_NEEDED
        sr.complete_auth_token(s_res.context, s_input_buffers)


@pytest.mark.skipif(os.name != "nt", reason="sspi-rs does not play well in this scenario")
def test_sec_exchange_allowed_package_list() -> None:
    spn = f"host/{socket.gethostname()}"
    buffer = bytearray(4096)

    auth_data = sr.WinNTAuthIdentity(package_list="NTLM")
    c_cred = sr.acquire_credentials_handle(
        None,
        "Negotiate",
        sr.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_data,
    )[0]
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sr.initialize_security_context(
        credential=c_cred,
        context=None,
        target_name=spn,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=None,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sr.InitializeContextResult)
    assert isinstance(c_res.context, sr.CtxtHandle)
    assert isinstance(c_res.attributes, sr.IscRet)
    assert isinstance(c_res.expiry, int)
    assert isinstance(c_res.status, sr.NtStatus)
    assert c_res.status == sr.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    s_cred = sr.acquire_credentials_handle(
        None,
        "Negotiate",
        sr.CredentialUse.SECPKG_CRED_INBOUND,
    )[0]
    s_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(c_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sr.accept_security_context(
        credential=s_cred,
        context=None,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sr.AcceptContextResult)
    assert isinstance(s_res.context, sr.CtxtHandle)
    assert isinstance(s_res.attributes, sr.AscRet)
    assert isinstance(s_res.expiry, int)
    assert isinstance(s_res.status, sr.NtStatus)
    assert s_res.status == sr.NtStatus.SEC_I_CONTINUE_NEEDED

    assert s_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert s_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    c_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(s_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sr.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sr.InitializeContextResult)
    assert isinstance(c_res.context, sr.CtxtHandle)
    assert isinstance(c_res.attributes, sr.IscRet)
    assert isinstance(c_res.expiry, int)
    assert isinstance(c_res.status, sr.NtStatus)
    assert c_res.status == sr.NtStatus.SEC_E_OK

    assert c_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    s_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(c_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sr.accept_security_context(
        credential=s_cred,
        context=s_res.context,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=None,
    )
    assert isinstance(s_res, sr.AcceptContextResult)
    assert isinstance(s_res.context, sr.CtxtHandle)
    assert isinstance(s_res.attributes, sr.AscRet)
    assert isinstance(s_res.expiry, int)
    assert isinstance(s_res.status, sr.NtStatus)
    assert s_res.status == sr.NtStatus.SEC_E_OK


@pytest.mark.skipif(os.name != "nt", reason="sspi-rs does not play well in this scenario")
def test_sec_exchange_with_negotiate_restricted_package() -> None:
    spn = f"host/{socket.gethostname()}"
    buffer = bytearray(4096)

    auth_data = sr.WinNTAuthIdentity(package_list="!Kerberos")
    c_cred = sr.acquire_credentials_handle(
        None,
        "Negotiate",
        sr.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_data,
    )[0]
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sr.initialize_security_context(
        credential=c_cred,
        context=None,
        target_name=spn,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=None,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sr.InitializeContextResult)
    assert isinstance(c_res.context, sr.CtxtHandle)
    assert isinstance(c_res.attributes, sr.IscRet)
    assert isinstance(c_res.expiry, int)
    assert isinstance(c_res.status, sr.NtStatus)
    assert c_res.status == sr.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    s_cred = sr.acquire_credentials_handle(
        None,
        "Negotiate",
        sr.CredentialUse.SECPKG_CRED_INBOUND,
    )[0]
    s_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(c_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sr.accept_security_context(
        credential=s_cred,
        context=None,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sr.AcceptContextResult)
    assert isinstance(s_res.context, sr.CtxtHandle)
    assert isinstance(s_res.attributes, sr.AscRet)
    assert isinstance(s_res.expiry, int)
    assert isinstance(s_res.status, sr.NtStatus)
    assert s_res.status == sr.NtStatus.SEC_I_CONTINUE_NEEDED

    assert s_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert s_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    c_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(s_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sr.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sr.InitializeContextResult)
    assert isinstance(c_res.context, sr.CtxtHandle)
    assert isinstance(c_res.attributes, sr.IscRet)
    assert isinstance(c_res.expiry, int)
    assert isinstance(c_res.status, sr.NtStatus)
    assert c_res.status == sr.NtStatus.SEC_I_CONTINUE_NEEDED

    assert c_output_buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert c_output_buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE

    s_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(c_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    s_res = sr.accept_security_context(
        credential=s_cred,
        context=s_res.context,
        input_buffers=s_input_buffers,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        output_buffers=s_output_buffers,
    )
    assert isinstance(s_res, sr.AcceptContextResult)
    assert isinstance(s_res.context, sr.CtxtHandle)
    assert isinstance(s_res.attributes, sr.AscRet)
    assert isinstance(s_res.expiry, int)
    assert isinstance(s_res.status, sr.NtStatus)
    assert s_res.status == sr.NtStatus.SEC_E_OK

    c_input_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(s_output_buffers[0].dangerous_get_view(), sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(buffer, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    c_res = sr.initialize_security_context(
        credential=c_cred,
        context=c_res.context,
        target_name=spn,
        context_req=0,
        target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
        input_buffers=c_input_buffers,
        output_buffers=c_output_buffers,
    )
    assert isinstance(c_res, sr.InitializeContextResult)
    assert isinstance(c_res.context, sr.CtxtHandle)
    assert isinstance(c_res.attributes, sr.IscRet)
    assert isinstance(c_res.expiry, int)
    assert isinstance(c_res.status, sr.NtStatus)
    assert s_res.status == sr.NtStatus.SEC_E_OK

    assert c_output_buffers[0].count == 0
    assert c_output_buffers[0].data == b""


# https://github.com/Devolutions/sspi-rs/issues/171
@pytest.mark.skipif(os.name != "nt", reason="sspi-rs does not support this scenario yet")
def test_fail_with_negotiate_and_no_available_packages() -> None:
    spn = f"host/{socket.gethostname()}"

    auth_data = sr.WinNTAuthIdentity(package_list="NegoExtender")
    c_cred = sr.acquire_credentials_handle(
        None,
        "Negotiate",
        sr.CredentialUse.SECPKG_CRED_OUTBOUND,
        auth_data=auth_data,
    )[0]
    c_output_buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(None, sr.SecBufferType.SECBUFFER_TOKEN),
        ]
    )
    with pytest.raises(sr.WindowsError) as e:
        sr.initialize_security_context(
            credential=c_cred,
            context=None,
            target_name=spn,
            context_req=sr.IscReq.ISC_REQ_ALLOCATE_MEMORY,
            target_data_rep=sr.TargetDataRep.SECURITY_NATIVE_DREP,
            input_buffers=None,
            output_buffers=c_output_buffers,
        )

    assert e.value.winerror == -2146893042  # SEC_E_NO_CREDENTIALS



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import dataclasses
import enum

from ._security_buffer cimport FreeContextBuffer
from ._text cimport WideCharString, wide_char_to_str
from ._win32_types cimport *


cdef extern from "python_sspi.h":
    unsigned int _SECPKG_FLAG_INTEGRITY "SECPKG_FLAG_INTEGRITY"
    unsigned int _SECPKG_FLAG_PRIVACY "SECPKG_FLAG_PRIVACY"
    unsigned int _SECPKG_FLAG_TOKEN_ONLY "SECPKG_FLAG_TOKEN_ONLY"
    unsigned int _SECPKG_FLAG_DATAGRAM "SECPKG_FLAG_DATAGRAM"
    unsigned int _SECPKG_FLAG_CONNECTION "SECPKG_FLAG_CONNECTION"
    unsigned int _SECPKG_FLAG_MULTI_REQUIRED "SECPKG_FLAG_MULTI_REQUIRED"
    unsigned int _SECPKG_FLAG_CLIENT_ONLY "SECPKG_FLAG_CLIENT_ONLY"
    unsigned int _SECPKG_FLAG_EXTENDED_ERROR "SECPKG_FLAG_EXTENDED_ERROR"
    unsigned int _SECPKG_FLAG_IMPERSONATION "SECPKG_FLAG_IMPERSONATION"
    unsigned int _SECPKG_FLAG_ACCEPT_WIN32_NAME "SECPKG_FLAG_ACCEPT_WIN32_NAME"
    unsigned int _SECPKG_FLAG_STREAM "SECPKG_FLAG_STREAM"
    unsigned int _SECPKG_FLAG_NEGOTIABLE "SECPKG_FLAG_NEGOTIABLE"
    unsigned int _SECPKG_FLAG_GSS_COMPATIBLE "SECPKG_FLAG_GSS_COMPATIBLE"
    unsigned int _SECPKG_FLAG_LOGON "SECPKG_FLAG_LOGON"
    unsigned int _SECPKG_FLAG_ASCII_BUFFERS "SECPKG_FLAG_ASCII_BUFFERS"
    unsigned int _SECPKG_FLAG_FRAGMENT "SECPKG_FLAG_FRAGMENT"
    unsigned int _SECPKG_FLAG_MUTUAL_AUTH "SECPKG_FLAG_MUTUAL_AUTH"
    unsigned int _SECPKG_FLAG_DELEGATION "SECPKG_FLAG_DELEGATION"
    unsigned int _SECPKG_FLAG_READONLY_WITH_CHECKSUM "SECPKG_FLAG_READONLY_WITH_CHECKSUM"
    unsigned int _SECPKG_FLAG_RESTRICTED_TOKENS "SECPKG_FLAG_RESTRICTED_TOKENS"
    unsigned int _SECPKG_FLAG_NEGO_EXTENDER "SECPKG_FLAG_NEGO_EXTENDER"
    unsigned int _SECPKG_FLAG_NEGOTIABLE2 "SECPKG_FLAG_NEGOTIABLE2"
    unsigned int _SECPKG_FLAG_APPCONTAINER_PASSTHROUGH "SECPKG_FLAG_APPCONTAINER_PASSTHROUGH"
    unsigned int _SECPKG_FLAG_APPCONTAINER_CHECKS "SECPKG_FLAG_APPCONTAINER_CHECKS"
    unsigned int _SECPKG_FLAG_CREDENTIAL_ISOLATION_ENABLED "SECPKG_FLAG_CREDENTIAL_ISOLATION_ENABLED"
    unsigned int _SECPKG_FLAG_APPLY_LOOPBACK "SECPKG_FLAG_APPLY_LOOPBACK"

    # https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-enumeratesecuritypackagesw
    SECURITY_STATUS EnumerateSecurityPackagesW(
        unsigned int *pcPackages,
        PSecPkgInfoW  *ppPackageInfo
    ) nogil

class SecurityPackageCapability(enum.IntFlag):
    SECPKG_FLAG_INTEGRITY = _SECPKG_FLAG_INTEGRITY
    SECPKG_FLAG_PRIVACY = _SECPKG_FLAG_PRIVACY
    SECPKG_FLAG_TOKEN_ONLY = _SECPKG_FLAG_TOKEN_ONLY
    SECPKG_FLAG_DATAGRAM = _SECPKG_FLAG_DATAGRAM
    SECPKG_FLAG_CONNECTION = _SECPKG_FLAG_CONNECTION
    SECPKG_FLAG_MULTI_REQUIRED = _SECPKG_FLAG_MULTI_REQUIRED
    SECPKG_FLAG_CLIENT_ONLY = _SECPKG_FLAG_CLIENT_ONLY
    SECPKG_FLAG_EXTENDED_ERROR = _SECPKG_FLAG_EXTENDED_ERROR
    SECPKG_FLAG_IMPERSONATION = _SECPKG_FLAG_IMPERSONATION
    SECPKG_FLAG_ACCEPT_WIN32_NAME = _SECPKG_FLAG_ACCEPT_WIN32_NAME
    SECPKG_FLAG_STREAM = _SECPKG_FLAG_STREAM
    SECPKG_FLAG_NEGOTIABLE = _SECPKG_FLAG_NEGOTIABLE
    SECPKG_FLAG_GSS_COMPATIBLE = _SECPKG_FLAG_GSS_COMPATIBLE
    SECPKG_FLAG_LOGON = _SECPKG_FLAG_LOGON
    SECPKG_FLAG_ASCII_BUFFERS = _SECPKG_FLAG_ASCII_BUFFERS
    SECPKG_FLAG_FRAGMENT = _SECPKG_FLAG_FRAGMENT
    SECPKG_FLAG_MUTUAL_AUTH = _SECPKG_FLAG_MUTUAL_AUTH
    SECPKG_FLAG_DELEGATION = _SECPKG_FLAG_DELEGATION
    SECPKG_FLAG_READONLY_WITH_CHECKSUM = _SECPKG_FLAG_READONLY_WITH_CHECKSUM
    SECPKG_FLAG_RESTRICTED_TOKENS = _SECPKG_FLAG_RESTRICTED_TOKENS
    SECPKG_FLAG_NEGO_EXTENDER = _SECPKG_FLAG_NEGO_EXTENDER
    SECPKG_FLAG_NEGOTIABLE2 = _SECPKG_FLAG_NEGOTIABLE2
    SECPKG_FLAG_APPCONTAINER_PASSTHROUGH = _SECPKG_FLAG_APPCONTAINER_PASSTHROUGH
    SECPKG_FLAG_APPCONTAINER_CHECKS = _SECPKG_FLAG_APPCONTAINER_CHECKS
    SECPKG_FLAG_CREDENTIAL_ISOLATION_ENABLED = _SECPKG_FLAG_CREDENTIAL_ISOLATION_ENABLED
    SECPKG_FLAG_APPLY_LOOPBACK = _SECPKG_FLAG_APPLY_LOOPBACK

@dataclasses.dataclass
class SecPkgInfo:
    capabilities: SecurityPackageCapability
    version: int
    rpcid: int
    max_token: int
    name: str
    comment: str

def enumerate_security_packages() -> list[SecPkgInfo]:
    cdef unsigned int num_packages = 0
    cdef PSecPkgInfoW packages = NULL

    with nogil:
        res = EnumerateSecurityPackagesW(&num_packages, &packages)

    if res:
        PyErr_SetFromWindowsErr(res)

    result = []
    try:
        for idx in range(num_packages):
            raw = packages[idx]
            info = SecPkgInfo(
                capabilities=SecurityPackageCapability(raw.fCapabilities),
                version=raw.wVersion,
                rpcid=raw.wRPCID,
                max_token=raw.cbMaxToken,
                name=wide_char_to_str(raw.Name),
                comment=wide_char_to_str(raw.Comment),
            )
            result.append(info)

    finally:
        FreeContextBuffer(packages)

    return result



# Copyright: (c) 2023, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from cpython.exc cimport PyErr_SetFromWindowsErr
from libc.stddef cimport size_t


cdef extern from "python_sspi.h":
    # PyErr_SetFromWindowsErr is Windows only, need a shim for Linux
    """
    #if defined(PYSSPI_IS_LINUX)
    PyObject *WinError;

    PyObject *PyErr_SetFromWindowsErr(int ierr)
    {
        if (WinError == NULL)
        {
            PyObject *winerror = PyImport_ImportModule("sspi.raw._winerror");
            if (winerror == NULL)
            {
                PyErr_SetString(PyExc_RuntimeError, "Failed to import custom WindowsError.");
                return NULL;
            }
            else
            {
                WinError = PyObject_GetAttrString(winerror, "WindowsError");
                Py_XDECREF(winerror);
            }
        }

        PyObject * err_obj = PyLong_FromLong(ierr);
        PyErr_SetObject(WinError, err_obj);
        return NULL;
    }
    #endif
    """

    ctypedef void *PVOID;

    ctypedef unsigned short USHORT

    ctypedef int LONG
    ctypedef unsigned int ULONG
    ctypedef LONG SECURITY_STATUS

    ctypedef unsigned short WCHAR
    ctypedef WCHAR *LPWSTR
    ctypedef WCHAR SEC_WCHAR;

    ctypedef size_t ULONG_PTR

    cdef struct _SecHandle:
        ULONG_PTR dwLower
        ULONG_PTR dwUpper
    ctypedef _SecHandle SecHandle
    ctypedef SecHandle *PSecHandle

    ctypedef SecHandle _CredHandle "CredHandle"
    ctypedef PSecHandle PCredHandle

    ctypedef SecHandle _CtxtHandle "CtxtHandle"
    ctypedef PSecHandle PCtxtHandle

    cdef struct _SECURITY_INTEGER:
        unsigned int LowPart
        int          HighPart
    ctypedef _SECURITY_INTEGER SECURITY_INTEGER
    ctypedef SECURITY_INTEGER *PSECURITY_INTEGER
    ctypedef SECURITY_INTEGER TimeStamp
    ctypedef SECURITY_INTEGER *PTimeStamp



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import sspi.raw as sr


def test_enumerate_security_packages() -> None:
    actual = sr.enumerate_security_packages()
    assert isinstance(actual, list)
    for sec_pkg in actual:
        assert isinstance(sec_pkg, sr.SecPkgInfo)
        assert isinstance(sec_pkg.capabilities, sr.SecurityPackageCapability)
        assert isinstance(sec_pkg.name, str)
        assert sec_pkg.name
        assert isinstance(sec_pkg.comment, str)



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import pytest

import sspi.raw as sr


def test_empty_sec_buffer_desc() -> None:
    buffers = sr.SecBufferDesc([])
    assert len(buffers) == 0
    assert list(iter(buffers)) == []
    assert buffers.version == 0


def test_sec_buffer_desc_version() -> None:
    buffers = sr.SecBufferDesc([], version=1)
    assert buffers.version == 1


def test_sec_buffer_desc_indexing() -> None:
    buffers = sr.SecBufferDesc(
        [
            sr.SecBuffer(bytearray(b"1"), sr.SecBufferType.SECBUFFER_DATA, sr.SecBufferFlags.SECBUFFER_READONLY),
            sr.SecBuffer(bytearray(b"2"), sr.SecBufferType.SECBUFFER_TOKEN),
            sr.SecBuffer(None, sr.SecBufferType.SECBUFFER_PADDING),
        ]
    )

    assert len(buffers) == 3

    assert isinstance(buffers[0], sr.SecBuffer)
    assert buffers[0].buffer_type == sr.SecBufferType.SECBUFFER_DATA
    assert buffers[0].buffer_flags == sr.SecBufferFlags.SECBUFFER_READONLY
    assert buffers[0].count == 1
    assert buffers[0].data == b"1"

    assert isinstance(buffers[1], sr.SecBuffer)
    assert buffers[1].buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert buffers[1].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE
    assert buffers[1].count == 1
    assert buffers[1].data == b"2"

    assert isinstance(buffers[2], sr.SecBuffer)
    assert buffers[2].buffer_type == sr.SecBufferType.SECBUFFER_PADDING
    assert buffers[2].buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE
    assert buffers[2].count == 0
    assert buffers[2].data == b""

    assert buffers[-1] == buffers[2]
    assert buffers[-2] == buffers[1]
    assert buffers[-3] == buffers[0]

    with pytest.raises(IndexError):
        buffers[3]

    with pytest.raises(IndexError):
        buffers[-4]


def test_sec_buffer_bytearray() -> None:
    data = bytearray(b"data")
    buffer = sr.SecBuffer(data, sr.SecBufferType.SECBUFFER_DATA)

    assert buffer.count == 4
    assert buffer.data == b"data"
    assert buffer.buffer_type == sr.SecBufferType.SECBUFFER_DATA
    assert buffer.buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE
    assert repr(buffer) == "SecBuffer(data=bytearray(b'data'), buffer_type=1, buffer_flags=0)"
    assert str(buffer) == "SECBUFFER_DATA"
    assert bytes(buffer.dangerous_get_view()) == b"data"


def test_sec_buffer_memoryview() -> None:
    data = bytearray(b"data")
    buffer = sr.SecBuffer(
        memoryview(data),
        sr.SecBufferType.SECBUFFER_TOKEN,
        sr.SecBufferFlags.SECBUFFER_READONLY_WITH_CHECKSUM,
    )

    assert buffer.count == 4
    assert buffer.data == b"data"
    assert buffer.buffer_type == sr.SecBufferType.SECBUFFER_TOKEN
    assert buffer.buffer_flags == sr.SecBufferFlags.SECBUFFER_READONLY_WITH_CHECKSUM
    assert repr(buffer) == "SecBuffer(data=bytearray(b'data'), buffer_type=2, buffer_flags=268435456)"
    assert str(buffer) == "SECBUFFER_TOKEN|SECBUFFER_READONLY_WITH_CHECKSUM"
    assert bytes(buffer.dangerous_get_view()) == b"data"


def test_sec_buffer_empty() -> None:
    data = bytearray(b"")
    buffer = sr.SecBuffer(data, sr.SecBufferType.SECBUFFER_DATA)

    assert buffer.count == 0
    assert buffer.data == b""
    assert buffer.buffer_type == sr.SecBufferType.SECBUFFER_DATA
    assert buffer.buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE
    assert repr(buffer) == "SecBuffer(data=bytearray(b''), buffer_type=1, buffer_flags=0)"
    assert str(buffer) == "SECBUFFER_DATA"
    assert bytes(buffer.dangerous_get_view()) == b""


def test_sec_buffer_none() -> None:
    buffer = sr.SecBuffer(None, sr.SecBufferType.SECBUFFER_DATA)

    assert buffer.count == 0
    assert buffer.data == b""
    assert buffer.buffer_type == sr.SecBufferType.SECBUFFER_DATA
    assert buffer.buffer_flags == sr.SecBufferFlags.SECBUFFER_NONE
    assert repr(buffer) == "SecBuffer(data=bytearray(b''), buffer_type=1, buffer_flags=0)"
    assert str(buffer) == "SECBUFFER_DATA"
    assert bytes(buffer.dangerous_get_view()) == b""


def test_sec_channel_bindings_nothing() -> None:
    buffer = sr.SecChannelBindings()
    assert (
        repr(buffer)
        == "SecChannelBindings(initiator_addr_type=0, initiator_addr=None, acceptor_addr_type=0, acceptor_addr=None, application_data=None)"
    )
    assert buffer.initiator_addr_type == 0
    assert buffer.initiator_addr is None
    assert buffer.acceptor_addr_type == 0
    assert buffer.acceptor_addr is None
    assert buffer.application_data is None

    data = buffer.get_sec_buffer_copy().data
    assert data == b"\x00" * 32


def test_sec_channel_bindings_initiator() -> None:
    buffer = sr.SecChannelBindings(initiator_addr_type=1, initiator_addr=b"12\x003")
    assert (
        repr(buffer)
        == "SecChannelBindings(initiator_addr_type=1, initiator_addr=b'12\\x003', acceptor_addr_type=0, acceptor_addr=None, application_data=None)"
    )
    assert buffer.initiator_addr_type == 1
    assert buffer.initiator_addr == b"12\x003"
    assert buffer.acceptor_addr_type == 0
    assert buffer.acceptor_addr is None
    assert buffer.application_data is None

    data = buffer.get_sec_buffer_copy().data
    assert data == b"".join(
        [
            b"\x01\x00\x00\x00\x04\x00\x00\x00\x20\x00\x00\x00",
            b"\x00" * 20,
            b"12\x003",
        ]
    )


def test_sec_channel_bindings_acceptor() -> None:
    buffer = sr.SecChannelBindings(acceptor_addr_type=1, acceptor_addr=b"12\x003")
    assert (
        repr(buffer)
        == "SecChannelBindings(initiator_addr_type=0, initiator_addr=None, acceptor_addr_type=1, acceptor_addr=b'12\\x003', application_data=None)"
    )
    assert buffer.initiator_addr_type == 0
    assert buffer.initiator_addr is None
    assert buffer.acceptor_addr_type == 1
    assert buffer.acceptor_addr == b"12\x003"
    assert buffer.application_data is None

    data = buffer.get_sec_buffer_copy().data
    assert data == b"".join(
        [
            b"\x00" * 12,
            b"\x01\x00\x00\x00\x04\x00\x00\x00\x20\x00\x00\x00",
            b"\x00" * 8,
            b"12\x003",
        ]
    )


def test_sec_channel_bindings_appdata() -> None:
    buffer = sr.SecChannelBindings(application_data=b"12\x003")
    assert (
        repr(buffer)
        == "SecChannelBindings(initiator_addr_type=0, initiator_addr=None, acceptor_addr_type=0, acceptor_addr=None, application_data=b'12\\x003')"
    )
    assert buffer.initiator_addr_type == 0
    assert buffer.initiator_addr is None
    assert buffer.acceptor_addr_type == 0
    assert buffer.acceptor_addr is None
    assert buffer.application_data == b"12\x003"

    data = buffer.get_sec_buffer_copy().data
    assert data == b"".join(
        [
            b"\x00" * 24,
            b"\x04\x00\x00\x00\x20\x00\x00\x0012\x003",
        ]
    )


def test_sec_channel_bindings_all() -> None:
    buffer = sr.SecChannelBindings(
        initiator_addr_type=1,
        initiator_addr=b"1",
        acceptor_addr_type=2,
        acceptor_addr=b"2",
        application_data=b"3",
    )
    assert (
        repr(buffer)
        == "SecChannelBindings(initiator_addr_type=1, initiator_addr=b'1', acceptor_addr_type=2, acceptor_addr=b'2', application_data=b'3')"
    )
    assert buffer.initiator_addr_type == 1
    assert buffer.initiator_addr == b"1"
    assert buffer.acceptor_addr_type == 2
    assert buffer.acceptor_addr == b"2"
    assert buffer.application_data == b"3"

    data = buffer.get_sec_buffer_copy().data
    assert data == b"".join(
        [
            b"\x01\x00\x00\x00\x01\x00\x00\x00\x20\x00\x00\x00",
            b"\x02\x00\x00\x00\x01\x00\x00\x00\x21\x00\x00\x00",
            b"\x01\x00\x00\x00\x22\x00\x00\x00",
            b"123",
        ]
    )


def test_sec_channel_bindings_dangerous_get_buffer() -> None:
    channel_bindings = sr.SecChannelBindings(application_data=b"12\x003")
    expected = b"".join(
        [
            b"\x00" * 24,
            b"\x04\x00\x00\x00\x20\x00\x00\x0012\x003",
        ]
    )

    safe_buffer = channel_bindings.get_sec_buffer_copy()
    unsafe_buffer = channel_bindings.dangerous_get_sec_buffer()

    safe_buffer.data == expected
    unsafe_buffer.data == expected

    # Modifying the raw bytes should reflect back on the channel_bindings val
    unsafe_buffer.dangerous_get_view()[0] = b"\x01"  # type: ignore[call-overload]
    assert channel_bindings.initiator_addr_type == 1

    # It should not affect the safe buffer created with a copy
    assert safe_buffer.data == expected
    assert unsafe_buffer.data == b"".join(
        [
            b"\x01",
            b"\x00" * 23,
            b"\x04\x00\x00\x00\x20\x00\x00\x0012\x003",
        ]
    )

    # The safe buffer should still be valid beyond the life of the object
    del channel_bindings
    safe_buffer.data == expected



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import enum
import typing as t

class CredentialUse(enum.IntFlag):
    """CredHandle Usage Flags for :meth:`acquire_credentials_handle`."""

    SECPKG_CRED_INBOUND = ...
    """Validate an incoming server credential."""

    SECPKG_CRED_OUTBOUND = ...
    """Allow a local client credential to prepare an outgoing token."""

    SECPKG_CRED_BOTH = ...
    """Validate an incoming credential or use a local credential to prepare an outgoing token."""

    SECPKG_CRED_AUTOLOGON_RESTRICTED = ...
    """Do not use the default logon credentials or credentials from CredHandle Manager"""

    SECPKG_CRED_PROCESS_POLICY_ONLY = ...
    """Process server policy."""

class WinNTAuthFlags(enum.IntFlag):
    SEC_WINNT_AUTH_IDENTITY_ANSI = ...
    """Credentials are in ANSI form."""
    SEC_WINNT_AUTH_IDENTITY_UNICODE = ...
    """Credentials are in Unicode form."""
    SEC_WINNT_AUTH_IDENTITY_MARSHALLED = ...
    """All data is in one buffer."""
    SEC_WINNT_AUTH_IDENTITY_ONLY = ...
    """Used with Kerberos to specify the cred are for identity only."""
    SEC_WINNT_AUTH_IDENTITY_FLAGS_PROCESS_ENCRYPTED = ...
    SEC_WINNT_AUTH_IDENTITY_FLAGS_SYSTEM_PROTECTED = ...
    SEC_WINNT_AUTH_IDENTITY_FLAGS_USER_PROTECTED = ...
    SEC_WINNT_AUTH_IDENTITY_FLAGS_SYSTEM_ENCRYPTED = ...
    SEC_WINNT_AUTH_IDENTITY_FLAGS_RESERVED = ...
    SEC_WINNT_AUTH_IDENTITY_FLAGS_NULL_USER = ...
    SEC_WINNT_AUTH_IDENTITY_FLAGS_NULL_DOMAIN = ...
    SEC_WINNT_AUTH_IDENTITY_FLAGS_ID_PROVIDER = ...

class CredHandle:
    """A credential handle.

    This contains the credential handle that was created with
    :meth:`acquire_credentials_handle`. Once no longer referenced, the handle
    will be freed internally, closing the credential handle.
    """

class AuthIdentity:
    """Base class for :meth:`acquire_credentials_handle` auth_data."""

class WinNTAuthIdentity(AuthIdentity):
    """Username and password identity information.

    This is an authentication data object that contains the username and
    password information for a user. It can be used with the ``auth_data``
    kwarg on :meth:`acquire_credentials_handle` to acquire a credential with
    an explicit username/password.

    If using the Netlogon form ``DOMAIN\\username``, the ``username`` arg is
    ``username`` and the ``domain`` arg is ``DOMAIN``. If using the UPN form
    ``username@DOMAIN.COM``, the ``username`` is that UPN value while ``domain``
    is None.

    The ``package_list`` kwarg can be used to restrict what security packages
    the ``Negotiate`` provider can use with this credential. For example to
    use the ``Negotiate`` provider but disable NTLM authentication the value
    ``!ntlm`` can be specified.

    The only supported flag is ``SEC_WINNT_AUTH_IDENTITY_ONLY``, setting any
    other value will have an unknown outcome.

    Args:
        username: The username.
        domain: The domain of the user.
        password: The user's password.
        flags: Custom flags associated with this identity.
        package_list: Comma separated list of names of security packages that
            are available to the Negotiate package.
    """

    def __init__(
        self,
        username: str | None = None,
        domain: str | None = None,
        password: str | None = None,
        *,
        flags: WinNTAuthFlags | int = 0,
        package_list: str | None = None,
    ) -> None: ...
    @property
    def username(self) -> str | None:
        """The identity username."""
    @property
    def domain(self) -> str | None:
        """The identity domain."""
    @property
    def password(self) -> str | None:
        """The identity password."""
    @property
    def flags(self) -> WinNTAuthFlags:
        """Custom flags associated with the identity."""
    @property
    def package_list(self) -> str | None:
        """A comma-separated list of security packages that are available to the Negotiate provider."""

class AcquireCredentialsResult(t.NamedTuple):
    """The acquire credentials handle result."""

    credential: CredHandle
    """
    The generated context to use for subsequent operations on this context.
    This context should be passed as the `context` arg on any remaining calls
    to :meth:`accept_security_context`.
    """
    expiry: int
    """The time at which the credential expires as a FILETIME value."""

def acquire_credentials_handle(
    principal: str | None,
    package: str,
    credential_use: CredentialUse | int,
    *,
    auth_data: AuthIdentity | None = None,
) -> AcquireCredentialsResult:
    """Acquires a handle to a credential.

    This function can be used to acquire a handle to a preexisting credential
    or to a new credential specified by this function. The credential can be
    used with functions like :meth:`initialize_sec_context` or
    :meth:`accept_sec_context` to perform the authentication steps.

    Currently the only authentication data value supported is
    :class:`WinNTAuthIdentity`.

    This wraps the `AcquireCredentialsHandle`_ Win32 function.

    Args:
        principal: Name of the principal whose credentials the handle will
            reference. The handling of this value is implemented in the
            security package requested.
        package: The security package name for which these credentials will be
            used.
        credential_use: Indicates how these credentials will be used.
        auth_data: Optional package specific authentication data.

    Returns:
        AcquireCredentialsResult: The acquired credential object and expiry.

    Raises:
        WindowsError: If the function failed.

    .. _AcquireCredentialsHandle:
        https://learn.microsoft.com/en-us/windows/win32/secauthn/acquirecredentialshandle--general
    """

def _replace_cred_handle(
    src: CredHandle,
    dst: CredHandle,
) -> None:
    """Internal use only."""
